from pathlib import Path

import pytest

from avrae_ls.api import CharacterAPI
from avrae_ls.completions import hover_for_position
from avrae_ls.config import AvraeLSConfig, VarSources
from avrae_ls.context import ContextData, ContextBuilder, GVarResolver


def _ctx_with_vars() -> tuple[ContextData, GVarResolver]:
    ctx = ContextData(vars=VarSources.from_data({"cvars": {"foo": 123}}))
    resolver = GVarResolver(AvraeLSConfig.default(Path(".")))
    resolver.reset({"g1": "hello"})
    return ctx, resolver


def test_hover_shows_var_value_and_type():
    ctx_data, resolver = _ctx_with_vars()
    hover = hover_for_position("foo\n", 0, 1, {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "foo" in text
    assert "int" in text
    assert "123" in text


def test_hover_shows_local_constant_value():
    ctx_data = ContextData()
    resolver = GVarResolver(AvraeLSConfig.default(Path(".")))
    hover = hover_for_position("x = 3\n", 0, 0, {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "x" in text
    assert "int" in text
    assert "3" in text


def test_hover_handles_attribute_inside_call():
    ctx_data, resolver = _ctx_with_vars()
    code = "res = character().attacks\n"
    hover = hover_for_position(code, 0, code.index("attacks") + 2, {}, ctx_data, resolver)
    assert hover is not None
    assert "character().attacks" in hover.contents.value


def test_hover_resolves_attribute_binding_value():
    config = AvraeLSConfig.default(Path("."))
    builder = ContextBuilder(config)
    ctx_data = builder.build()
    resolver = builder.gvar_resolver
    code = "x = character()\ny = character().name\n"
    hover = hover_for_position(code, 1, 0, {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "y" in text
    assert "str" in text
    assert "Aelar" in text


def test_hover_resolves_attribute_via_variable():
    config = AvraeLSConfig.default(Path("."))
    builder = ContextBuilder(config)
    ctx_data = builder.build()
    resolver = builder.gvar_resolver
    code = "x = character()\ny = x.name\n"
    hover = hover_for_position(code, 1, 0, {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "y" in text
    assert "str" in text
    assert "Aelar" in text


def test_hover_resolves_method_call_on_binding():
    config = AvraeLSConfig.default(Path("."))
    builder = ContextBuilder(config)
    ctx_data = builder.build()
    resolver = builder.gvar_resolver
    code = "x = character()\ny = x.levels\nz = y.get('Fighter')\n"
    hover = hover_for_position(code, 2, 0, {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "z" in text
    assert "int" in text
    assert "3" in text


def test_hover_does_not_call_unsafe_methods(monkeypatch: pytest.MonkeyPatch):
    config = AvraeLSConfig.default(Path("."))
    builder = ContextBuilder(config)
    ctx_data = builder.build()
    resolver = builder.gvar_resolver

    def _boom(*_args, **_kwargs):
        raise AssertionError("unsafe method should not be called")

    monkeypatch.setattr(CharacterAPI, "set_cvar", _boom)
    code = "x = character()\ny = x.set_cvar('foo', 'bar')\n"
    hover = hover_for_position(code, 1, 0, {}, ctx_data, resolver)
    assert hover is None


def test_hover_for_loop_target():
    ctx_data = ContextData()
    resolver = GVarResolver(AvraeLSConfig.default(Path(".")))
    code = "for i in range(3):\n    return i\n"
    hover = hover_for_position(code, 1, code.index("i", code.index("return")), {}, ctx_data, resolver)
    assert hover is not None
    text = hover.contents.value
    assert "i" in text
    assert "int" in text



import collections.abc
import operator as op
from collections import UserList, UserString

from .exceptions import *
from .string import JoinProxy, PRINTF_TEMPLATE_RE, TranslateTableProxy
from .versions import PY_39

__all__ = ("safe_list", "safe_dict", "safe_set", "safe_str", "approx_len_of")

_sentinel = object()


# ---- size helper ----
def approx_len_of(obj, visited=None):
    """Gets the approximate size of an object (including recursive objects)."""
    if isinstance(obj, (str, bytes, UserString)):
        return len(obj)

    if hasattr(obj, "__approx_len__"):
        return obj.__approx_len__

    if visited is None:
        visited = [obj]

    size = op.length_hint(obj)

    if isinstance(obj, dict):
        obj = obj.items()

    try:
        obj_iter = iter(obj)
    except TypeError:  # object is not iterable
        pass
    else:
        for child in obj_iter:
            if child in visited:
                continue
            size += approx_len_of(child, visited)
            visited.append(child)

    try:
        setattr(obj, "__approx_len__", size)
    except (AttributeError, TypeError):
        pass

    return size


# ---- types ----
# each function is a function that returns a class based on Draconic config
# ... look, it works
def safe_list(config):
    class SafeList(UserList):  # extends UserList so that [x] * y returns a SafeList, not a list
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__approx_len__ = approx_len_of(self)

        def append(self, obj):
            if approx_len_of(self) + 1 > config.max_const_len:
                _raise_in_context(IterableTooLong, "This list is too long")
            super().append(obj)
            self.__approx_len__ += 1

        def extend(self, iterable):
            other_len = approx_len_of(iterable)
            if approx_len_of(self) + other_len > config.max_const_len:
                _raise_in_context(IterableTooLong, "This list is too long")
            super().extend(iterable)
            self.__approx_len__ += other_len

        def pop(self, i=-1):
            retval = super().pop(i)
            self.__approx_len__ -= 1
            return retval

        def remove(self, item):
            super().remove(item)
            self.__approx_len__ -= 1

        def clear(self):
            super().clear()
            self.__approx_len__ = 0

        def __mul__(self, n):
            # to prevent the recalculation of the length on list mult we manually set a new instance's
            # data and approx len (JIRA-54)
            new = SafeList()
            new.data = self.data * n
            new.__approx_len__ = self.__approx_len__ * n
            return new

    return SafeList


def safe_set(config):
    class SafeSet(set):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__approx_len__ = approx_len_of(self)

        def union(self, *s):
            if approx_len_of(self) + sum(approx_len_of(other) for other in s) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This set is too large")
            return SafeSet(super().union(*s))

        def __or__(self, other):
            return self.union(other)

        def intersection(self, *s):
            if any(approx_len_of(other) > config.max_const_len for other in s):
                _raise_in_context(IterableTooLong, "This set is too large")
            return SafeSet(super().intersection(*s))

        def __and__(self, other):
            return self.intersection(other)

        def symmetric_difference(self, *s):
            if approx_len_of(self) + sum(approx_len_of(other) for other in s) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This set is too large")
            return SafeSet(super().symmetric_difference(*s))

        def __xor__(self, other):
            return self.symmetric_difference(other)

        # difference not reimplemented as it cannot grow the set and has no cheap approximation for len

        def update(self, *s):
            other_lens = sum(approx_len_of(other) for other in s)
            if approx_len_of(self) + other_lens > config.max_const_len:
                _raise_in_context(IterableTooLong, "This set is too large")
            super().update(*s)
            self.__approx_len__ += other_lens

        def intersection_update(self, *s):
            if any(approx_len_of(other) > config.max_const_len for other in s):
                _raise_in_context(IterableTooLong, "This set is too large")
            super().intersection_update(*s)
            self.__approx_len__ = min(self.__approx_len__, *(approx_len_of(other) for other in s))

        def symmetric_difference_update(self, s):
            total_approx = approx_len_of(self) + approx_len_of(s)
            if total_approx > config.max_const_len:
                _raise_in_context(IterableTooLong, "This set is too large")
            super().symmetric_difference_update(s)
            self.__approx_len__ = total_approx

        # difference_update not reimplemented as it cannot grow the set and has no cheap approximation for len

        def add(self, element):
            if approx_len_of(self) + 1 > config.max_const_len:
                _raise_in_context(IterableTooLong, "This set is too large")
            super().add(element)
            self.__approx_len__ += 1

        def pop(self):
            retval = super().pop()
            self.__approx_len__ -= 1
            return retval

        def remove(self, element):
            super().remove(element)
            self.__approx_len__ -= 1

        # discard not reimplemented as the discarded element may not be a member

        def clear(self):
            super().clear()
            self.__approx_len__ = 0

    return SafeSet


def safe_dict(config):
    class SafeDict(dict):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.__approx_len__ = approx_len_of(self)

        def update(self, other_dict=None, **kvs):
            if other_dict is None:
                other_dict = {}

            other_lens = approx_len_of(other_dict) + approx_len_of(kvs)
            if approx_len_of(self) + other_lens > config.max_const_len:
                _raise_in_context(IterableTooLong, "This dict is too large")

            super().update(other_dict, **kvs)
            self.__approx_len__ += other_lens

        def __setitem__(self, key, value):
            other_len = approx_len_of(value)
            if approx_len_of(self) + other_len > config.max_const_len:
                _raise_in_context(IterableTooLong, "This dict is too large")
            self.__approx_len__ += other_len
            return super().__setitem__(key, value)

        def pop(self, k, default=_sentinel):
            if default is not _sentinel:
                retval = super().pop(k, default)
            else:
                retval = super().pop(k)
                self.__approx_len__ -= 1
            return retval

        def __delitem__(self, key):
            super().__delitem__(key)
            self.__approx_len__ -= 1

        def __getattr__(self, attr):
            try:
                return self[attr]
            except KeyError:
                raise AttributeError

        if PY_39:

            def __or__(self, other):
                if approx_len_of(self) + approx_len_of(other) > config.max_const_len:
                    _raise_in_context(IterableTooLong, "This dict is too large")

                return SafeDict(super().__or__(other))

    return SafeDict


_real_str = str


def safe_str(config):
    # noinspection PyShadowingBuiltins, PyPep8Naming
    # naming it SafeStr would break typeof backward compatibility :(
    class str(UserString, _real_str):
        def __init__(self, seq):
            if isinstance(seq, UserString):
                self.data = seq.data[:]
            elif isinstance(seq, _real_str):
                self.data = seq
            else:
                self.data = _real_str(seq)

        def center(self, width, *args):
            if width > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().center(width, *args)

        def encode(self, *_, **__):
            _raise_in_context(FeatureNotAvailable, "This method is not allowed")

        def expandtabs(self, tabsize=8):
            if self.count("\t") * tabsize > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().expandtabs(tabsize)

        def format(self, *args, **kwargs):
            _raise_in_context(FeatureNotAvailable, "This method is not allowed")

        def format_map(self, mapping):
            _raise_in_context(FeatureNotAvailable, "This method is not allowed")

        def join(self, seq):
            full_seq = list(seq)  # consume the entire iterator so we can do length checking
            i = JoinProxy(self.__class__, full_seq)  # proxy it so that .join gets the right types
            if len(full_seq) * len(self) + approx_len_of(full_seq) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().join(i)

        def ljust(self, width, *args):
            if width > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().ljust(width, *args)

        @staticmethod
        def maketrans(*args):
            if len(args) == 1 and isinstance(args[0], dict):
                # str.maketrans expects a dict object and nothing else
                # So SafeDict needs to be cast to dict for the method to work
                return _real_str.maketrans(dict(args[0]))

            if sum(approx_len_of(a) for a in args) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This dict is too large")

            return _real_str.maketrans(*args)

        def replace(self, old, new, maxsplit=-1):
            if maxsplit > 0:
                n = maxsplit
            else:
                n = self.count(old)
            if n * (len(new) - len(old)) + len(self) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().replace(old, new, maxsplit)

        def rjust(self, width, *args):
            if width > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().rjust(width, *args)

        def translate(self, table):
            # this is kind of a disgusting way to check the worst-case length
            # and is an overestimate by a multiplicative factor of len(table)
            # but it is certainly an overestimate
            if approx_len_of(table) * len(self) > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            table_proxy = TranslateTableProxy(self.__class__, table)
            return super().translate(table_proxy)

        def zfill(self, width):
            if width > config.max_const_len:
                _raise_in_context(IterableTooLong, "This str is too large")
            return super().zfill(width)

        def __format__(self, format_spec):
            # format it using default str formatter
            return self.data.__format__(format_spec)

        def __mod__(self, values):
            new_len_bound = len(self)
            values_is_sequence = isinstance(values, collections.abc.Sequence)
            values_is_mapping = isinstance(values, collections.abc.Mapping)

            # validate that the template is safe (no massive widths/precisions)
            i = 0
            for match in PRINTF_TEMPLATE_RE.finditer(self.data):
                w = match.group("width")
                if w:
                    if w == "*":
                        _raise_in_context(FeatureNotAvailable, "Star precision in printf-style formatting not allowed")
                    else:
                        new_len_bound += int(w)

                p = match.group("precision")
                if p:
                    if p == "*":
                        _raise_in_context(FeatureNotAvailable, "Star precision in printf-style formatting not allowed")
                    else:
                        new_len_bound += int(p)

                mapping_key = match.group("mapping_key")
                if mapping_key is not None:  # '%(foo)s %(foo)s'
                    if not values_is_mapping:  # '%(foo)s' % 0
                        raise TypeError("format requires a mapping")
                    val = values[mapping_key]
                    new_len_bound += approx_len_of(val)
                elif values_is_sequence:  # '%s %s'
                    try:
                        val = values[i]
                    except IndexError:  # '%s %s' % [0]
                        raise TypeError("not enough arguments for format string")
                    new_len_bound += approx_len_of(val)
                elif i > 0:  # '%s %s' % 0
                    raise TypeError("not enough arguments for format string")
                else:  # '%s' % 0
                    new_len_bound += approx_len_of(values)

                if match.group("type") != "%":  # percent literals do not increase index
                    i += 1

                if new_len_bound > config.max_const_len:
                    _raise_in_context(IterableTooLong, "This str is too large")

            return _real_str.__mod__(self.data, values)

    return str



from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import Any, ClassVar, Iterable, Mapping, MutableMapping, Sequence

import d20

from .dice import RerollableStringifier

UNSET = object()


def _safe_int(value: Any, default: int = 0) -> int:
    try:
        return int(value)
    except Exception:
        return default


class _DirMixin:
    ATTRS: ClassVar[list[str]] = []
    METHODS: ClassVar[list[str]] = []

    def __dir__(self) -> list[str]:
        data_keys = list(getattr(self, "data", {}).keys())
        return sorted(set(self.ATTRS + self.METHODS + data_keys))

    def __post_init__(self) -> None:
        # Ensure draconic's approx_len cache is initialized to a numeric value
        # so nested objects don't trip TypeErrors when measured.
        if getattr(self, "__approx_len__", None) is None:
            self.__approx_len__ = 0


class SimpleRollResult(_DirMixin):
    ATTRS: ClassVar[list[str]] = ["dice", "total", "full", "result", "raw"]
    METHODS: ClassVar[list[str]] = ["consolidated"]

    def __init__(self, roll_result: d20.RollResult):
        self._roll = roll_result
        self.dice = d20.MarkdownStringifier().stringify(roll_result.expr.roll)
        self.total = roll_result.total
        self.full = str(roll_result)
        self.result = roll_result
        self.raw = roll_result.expr
        self.__approx_len__ = 0

    def consolidated(self) -> str:
        d20.utils.simplify_expr(self._roll.expr, ambig_inherit="left")
        return RerollableStringifier().stringify(self._roll.expr.roll)

    def __str__(self) -> str:
        return self.full


# === Context API ===
@dataclass
class GuildAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["name", "id"]
    METHODS: ClassVar[list[str]] = ["servsettings"]

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Guild"))

    @property
    def id(self) -> int | None:
        raw = self.data.get("id")
        return int(raw) if raw is not None else None

    def servsettings(self) -> Mapping[str, Any] | None:
        return self.data.get("servsettings")

    def __getitem__(self, item: str) -> Any:
        if isinstance(item, int):
            raise TypeError("CustomCounter indices must be strings (e.g., 'value', 'max').")
        return getattr(self, str(item))

    def __iter__(self):
        return iter(())

    def __len__(self) -> int:
        return 0


@dataclass
class CategoryAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["name", "id"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Category"))

    @property
    def id(self) -> int | None:
        raw = self.data.get("id")
        return int(raw) if raw is not None else None

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class ChannelAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["name", "id", "topic", "category", "parent"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "channel"))

    @property
    def id(self) -> int | None:
        raw = self.data.get("id")
        return int(raw) if raw is not None else None

    @property
    def topic(self) -> str | None:
        val = self.data.get("topic")
        return str(val) if val is not None else None

    @property
    def category(self) -> CategoryAPI | None:
        cat = self.data.get("category")
        return CategoryAPI(cat) if cat is not None else None

    @property
    def parent(self) -> ChannelAPI | None:
        parent = self.data.get("parent")
        return ChannelAPI(parent) if parent is not None else None

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class RoleAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["name", "id"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Role"))

    @property
    def id(self) -> int | None:
        raw = self.data.get("id")
        return int(raw) if raw is not None else None

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class AuthorAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["name", "id", "discriminator", "display_name", "roles"]
    METHODS: ClassVar[list[str]] = ["get_roles"]

    @property
    def name(self) -> str:
        return str(self.data.get("name", "User"))

    @property
    def id(self) -> int | None:
        raw = self.data.get("id")
        return int(raw) if raw is not None else None

    @property
    def discriminator(self) -> str:
        return str(self.data.get("discriminator", "0000"))

    @property
    def display_name(self) -> str:
        return str(self.data.get("display_name", self.name))

    def get_roles(self) -> list[RoleAPI]:
        roles = self.data.get("roles") or []
        return [RoleAPI(r) for r in roles]

    @property
    def roles(self) -> list[RoleAPI]:
        return self.get_roles()

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))

    def __str__(self) -> str:
        return f"{self.name}#{self.discriminator}"


@dataclass
class AliasContextAPI(_DirMixin):
    data: Mapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["guild", "channel", "author", "prefix", "alias", "message_id"]
    METHODS: ClassVar[list[str]] = []

    @property
    def guild(self) -> GuildAPI | None:
        guild_data = self.data.get("guild")
        return GuildAPI(guild_data) if guild_data is not None else None

    @property
    def channel(self) -> ChannelAPI | None:
        channel_data = self.data.get("channel")
        return ChannelAPI(channel_data) if channel_data is not None else None

    @property
    def author(self) -> AuthorAPI | None:
        author_data = self.data.get("author")
        return AuthorAPI(author_data) if author_data is not None else None

    @property
    def prefix(self) -> str | None:
        val = self.data.get("prefix")
        return str(val) if val is not None else None

    @property
    def alias(self) -> str | None:
        val = self.data.get("alias")
        return str(val) if val is not None else None

    @property
    def message_id(self) -> int | None:
        raw = self.data.get("message_id")
        return int(raw) if raw is not None else None

    def __getattr__(self, item: str) -> Any:
        return self.data.get(item)

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


# === StatBlock primitives ===
_SKILL_CANONICAL = {
    "acrobatics": "acrobatics",
    "animalhandling": "animalHandling",
    "arcana": "arcana",
    "athletics": "athletics",
    "deception": "deception",
    "history": "history",
    "initiative": "initiative",
    "insight": "insight",
    "intimidation": "intimidation",
    "investigation": "investigation",
    "medicine": "medicine",
    "nature": "nature",
    "perception": "perception",
    "performance": "performance",
    "persuasion": "persuasion",
    "religion": "religion",
    "sleightofhand": "sleightOfHand",
    "sleight_of_hand": "sleightOfHand",
    "stealth": "stealth",
    "survival": "survival",
    "strength": "strength",
    "dexterity": "dexterity",
    "constitution": "constitution",
    "intelligence": "intelligence",
    "wisdom": "wisdom",
    "charisma": "charisma",
}

_SKILL_ABILITIES = {
    "acrobatics": "dexterity",
    "animalHandling": "wisdom",
    "arcana": "intelligence",
    "athletics": "strength",
    "deception": "charisma",
    "history": "intelligence",
    "initiative": "dexterity",
    "insight": "wisdom",
    "intimidation": "charisma",
    "investigation": "intelligence",
    "medicine": "wisdom",
    "nature": "intelligence",
    "perception": "wisdom",
    "performance": "charisma",
    "persuasion": "charisma",
    "religion": "intelligence",
    "sleightOfHand": "dexterity",
    "stealth": "dexterity",
    "survival": "wisdom",
    "strength": "strength",
    "dexterity": "dexterity",
    "constitution": "constitution",
    "intelligence": "intelligence",
    "wisdom": "wisdom",
    "charisma": "charisma",
}


@dataclass
class AliasBaseStats(_DirMixin):
    data: Mapping[str, Any]
    prof_bonus_override: int | None = None
    ATTRS: ClassVar[list[str]] = [
        "prof_bonus",
        "strength",
        "dexterity",
        "constitution",
        "intelligence",
        "wisdom",
        "charisma",
    ]
    METHODS: ClassVar[list[str]] = ["get_mod", "get"]

    @property
    def prof_bonus(self) -> int:
        if self.prof_bonus_override is not None:
            return self.prof_bonus_override
        return _safe_int(self.data.get("prof_bonus"), 2)

    @property
    def strength(self) -> int:
        return _safe_int(self.data.get("strength"), 10)

    @property
    def dexterity(self) -> int:
        return _safe_int(self.data.get("dexterity"), 10)

    @property
    def constitution(self) -> int:
        return _safe_int(self.data.get("constitution"), 10)

    @property
    def intelligence(self) -> int:
        return _safe_int(self.data.get("intelligence"), 10)

    @property
    def wisdom(self) -> int:
        return _safe_int(self.data.get("wisdom"), 10)

    @property
    def charisma(self) -> int:
        return _safe_int(self.data.get("charisma"), 10)

    def get_mod(self, stat: str) -> int:
        stat_lower = str(stat).lower()
        lookup = {
            "str": self.strength,
            "dex": self.dexterity,
            "con": self.constitution,
            "int": self.intelligence,
            "wis": self.wisdom,
            "cha": self.charisma,
            "strength": self.strength,
            "dexterity": self.dexterity,
            "constitution": self.constitution,
            "intelligence": self.intelligence,
            "wisdom": self.wisdom,
            "charisma": self.charisma,
        }
        score = lookup.get(stat_lower, 10)
        return math.floor((score - 10) / 2)

    def get(self, stat: str) -> int:
        stat_lower = str(stat).lower()
        lookup = {
            "strength": self.strength,
            "dexterity": self.dexterity,
            "constitution": self.constitution,
            "intelligence": self.intelligence,
            "wisdom": self.wisdom,
            "charisma": self.charisma,
        }
        return lookup.get(stat_lower, 10)

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class AliasLevels(_DirMixin):
    data: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["total_level"]
    METHODS: ClassVar[list[str]] = ["get"]

    @property
    def total_level(self) -> int | float:
        total = 0
        for _, value in self.data.items():
            try:
                total += value
            except Exception:
                try:
                    total += int(value)
                except Exception:
                    continue
        return total

    def get(self, cls_name: str, default: int | float = 0) -> int | float:
        val = self.data.get(str(cls_name))
        if val is None:
            return default
        try:
            return val
        except Exception:
            return default

    def __iter__(self) -> Iterable[tuple[str, Any]]:
        return iter(self.data.items())

    def __getitem__(self, item: str) -> Any:
        return self.data[item]


@dataclass
class AliasAttack(_DirMixin):
    data: Mapping[str, Any]
    parent_statblock: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["name", "verb", "proper", "activation_type", "raw"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Attack"))

    @property
    def verb(self) -> str | None:
        val = self.data.get("verb")
        return str(val) if val is not None else None

    @property
    def proper(self) -> bool:
        return bool(self.data.get("proper", False))

    @property
    def activation_type(self) -> int | None:
        raw = self.data.get("activation_type")
        return int(raw) if raw is not None else None

    @property
    def raw(self) -> Mapping[str, Any]:
        return self.data.get("raw", self.data)

    def __str__(self) -> str:
        damage = self.raw.get("damage")
        verb = f" {self.verb}" if self.verb else ""
        if damage:
            return f"{self.name}{verb}: {damage}"
        return f"{self.name}{verb}".strip()

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class AliasAttackList(_DirMixin):
    attacks: Sequence[Mapping[str, Any]]
    parent_statblock: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = []
    METHODS: ClassVar[list[str]] = []

    def __iter__(self) -> Iterable[AliasAttack]:
        for atk in self.attacks:
            yield AliasAttack(atk, self.parent_statblock)

    def __getitem__(self, item: int | slice) -> Any:
        if isinstance(item, slice):
            return AliasAttackList(self.attacks[item], self.parent_statblock)
        return AliasAttack(self.attacks[item], self.parent_statblock)

    def __len__(self) -> int:
        return len(self.attacks)

    def __str__(self) -> str:
        return "\n".join(str(AliasAttack(atk, self.parent_statblock)) for atk in self.attacks)


@dataclass
class AliasSkill(_DirMixin):
    data: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["value", "prof", "bonus", "adv"]
    METHODS: ClassVar[list[str]] = ["d20"]

    @property
    def value(self) -> int:
        return _safe_int(self.data.get("value"), 0)

    @property
    def prof(self) -> float | int:
        raw = self.data.get("prof")
        try:
            return float(raw)
        except Exception:
            return 0

    @property
    def bonus(self) -> int:
        return _safe_int(self.data.get("bonus"), 0)

    @property
    def adv(self) -> bool | None:
        val = self.data.get("adv")
        if val is None:
            return None
        return bool(val)

    def d20(self, base_adv=None, reroll=None, min_val=None, mod_override=None) -> str:
        mod = mod_override if mod_override is not None else self.value
        adv_prefix = "2d20kh1" if base_adv else "2d20kl1" if base_adv is False else "1d20"
        suffix = f"+{mod}" if mod >= 0 else str(mod)
        parts = [adv_prefix + suffix]
        if reroll is not None:
            parts.append(f"(reroll {reroll})")
        if min_val is not None:
            parts.append(f"(min {min_val})")
        return " ".join(parts)

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f"<AliasSkill value={self.value} prof={self.prof} bonus={self.bonus} adv={self.adv}>"

    def __gt__(self, other: Any) -> bool:
        return self.value > other

    def __ge__(self, other: Any) -> bool:
        return self.value >= other

    def __eq__(self, other: Any) -> bool:
        return self.value == other

    def __le__(self, other: Any) -> bool:
        return self.value <= other

    def __lt__(self, other: Any) -> bool:
        return self.value < other


@dataclass
class AliasSkills(_DirMixin):
    data: Mapping[str, Any]
    prof_bonus: int
    abilities: Mapping[str, int]
    ATTRS: ClassVar[list[str]] = list(_SKILL_ABILITIES.keys())
    METHODS: ClassVar[list[str]] = []

    def __getattr__(self, item: str) -> AliasSkill:
        return self._get_skill(item)

    def __getitem__(self, item: str) -> AliasSkill:
        return self._get_skill(str(item))

    def _get_skill(self, name: str) -> AliasSkill:
        normalized = _SKILL_CANONICAL.get(name.lower().replace(" ", "").replace("_", ""), name)
        skill_data = self.data.get(normalized) or self.data.get(name) or {}
        if not skill_data:
            ability = _SKILL_ABILITIES.get(normalized)
            ability_mod = 0
            if ability:
                ability_score = self.abilities.get(ability, 10)
                ability_mod = math.floor((ability_score - 10) / 2)
            skill_data = {"value": ability_mod, "prof": 0, "bonus": 0, "adv": None}
        return AliasSkill(skill_data)

    def __iter__(self) -> Iterable[tuple[str, AliasSkill]]:
        for name in _SKILL_ABILITIES.keys():
            yield name, self._get_skill(name)

    def __str__(self) -> str:
        return ", ".join(f"{name}: {skill.value}" for name, skill in self)


@dataclass
class AliasSaves(_DirMixin):
    data: Mapping[str, Any]
    prof_bonus: int
    abilities: Mapping[str, int]
    ATTRS: ClassVar[list[str]] = []
    METHODS: ClassVar[list[str]] = ["get"]

    def get(self, base_stat: str) -> AliasSkill:
        normalized = base_stat.lower()
        raw = self.data.get(normalized) or self.data.get(base_stat) or {}
        if isinstance(raw, (int, float)):
            raw = {"value": raw}
        if not raw:
            ability_score = self.abilities.get(normalized, 10)
            raw = {"value": math.floor((ability_score - 10) / 2)}
        return AliasSkill(raw)

    def __iter__(self) -> Iterable[tuple[str, AliasSkill]]:
        for key in ("str", "dex", "con", "int", "wis", "cha"):
            yield key, self.get(key)

    def __str__(self) -> str:
        return ", ".join(f"{name}: {skill.value}" for name, skill in self)


@dataclass
class ResistanceEntry:
    dtype: str
    unless: set[str] = field(default_factory=set)
    only: set[str] = field(default_factory=set)


@dataclass
class AliasResistances(_DirMixin):
    data: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["resist", "vuln", "immune", "neutral"]
    METHODS: ClassVar[list[str]] = ["is_resistant", "is_immune", "is_vulnerable", "is_neutral"]

    @staticmethod
    def _entries(key: str, data: Mapping[str, Any]) -> list[ResistanceEntry]:
        entries = []
        for entry in data.get(key, []):
            entries.append(
                ResistanceEntry(
                    dtype=str(entry.get("dtype", "")),
                    unless=set(entry.get("unless", []) or []),
                    only=set(entry.get("only", []) or []),
                )
            )
        return entries

    @property
    def resist(self) -> list[ResistanceEntry]:
        return self._entries("resist", self.data)

    @property
    def vuln(self) -> list[ResistanceEntry]:
        return self._entries("vuln", self.data)

    @property
    def immune(self) -> list[ResistanceEntry]:
        return self._entries("immune", self.data)

    @property
    def neutral(self) -> list[ResistanceEntry]:
        return self._entries("neutral", self.data)

    def is_resistant(self, damage_type: str) -> bool:
        token = str(damage_type).lower()
        return self._matches(self.resist, token)

    def is_immune(self, damage_type: str) -> bool:
        token = str(damage_type).lower()
        return self._matches(self.immune, token)

    def is_vulnerable(self, damage_type: str) -> bool:
        token = str(damage_type).lower()
        return self._matches(self.vuln, token)

    def is_neutral(self, damage_type: str) -> bool:
        token = str(damage_type).lower()
        return self._matches(self.neutral, token)

    @staticmethod
    def _matches(entries: Iterable[ResistanceEntry], token: str) -> bool:
        for entry in entries:
            if entry.only and token not in entry.only:
                continue
            if entry.unless and token in entry.unless:
                continue
            if entry.dtype.lower() == token:
                return True
        return False


@dataclass
class AliasSpellbookSpell(_DirMixin):
    data: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["name", "dc", "sab", "mod", "prepared"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Spell"))

    @property
    def dc(self) -> int | None:
        raw = self.data.get("dc")
        return int(raw) if raw is not None else None

    @property
    def sab(self) -> int | None:
        raw = self.data.get("sab")
        return int(raw) if raw is not None else None

    @property
    def mod(self) -> int | None:
        raw = self.data.get("mod")
        return int(raw) if raw is not None else None

    @property
    def prepared(self) -> bool:
        return bool(self.data.get("prepared", True))

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))

    def __str__(self) -> str:
        return self.name


@dataclass
class AliasSpellbook(_DirMixin):
    data: MutableMapping[str, Any]
    _spells_cache: list[AliasSpellbookSpell] = field(default_factory=list, init=False)
    ATTRS: ClassVar[list[str]] = [
        "dc",
        "sab",
        "caster_level",
        "spell_mod",
        "spells",
        "pact_slot_level",
        "num_pact_slots",
        "max_pact_slots",
    ]
    METHODS: ClassVar[list[str]] = [
        "find",
        "slots_str",
        "get_max_slots",
        "get_slots",
        "set_slots",
        "use_slot",
        "reset_slots",
        "reset_pact_slots",
        "remaining_casts_of",
        "cast",
        "can_cast",
    ]

    @property
    def dc(self) -> int:
        return _safe_int(self.data.get("dc"), 10)

    @property
    def sab(self) -> int:
        return _safe_int(self.data.get("sab"), 0)

    @property
    def caster_level(self) -> int:
        return _safe_int(self.data.get("caster_level"), 0)

    @property
    def spell_mod(self) -> int:
        return _safe_int(self.data.get("spell_mod"), 0)

    @property
    def spells(self) -> list[AliasSpellbookSpell]:
        if not self._spells_cache:
            spell_list = self.data.get("spells") or []
            self._spells_cache = [AliasSpellbookSpell(s) for s in spell_list]
        return self._spells_cache

    @property
    def pact_slot_level(self) -> int | None:
        raw = self.data.get("pact_slot_level")
        return int(raw) if raw is not None else None

    @property
    def num_pact_slots(self) -> int | None:
        raw = self.data.get("num_pact_slots")
        return int(raw) if raw is not None else None

    @property
    def max_pact_slots(self) -> int | None:
        raw = self.data.get("max_pact_slots")
        return int(raw) if raw is not None else None

    def find(self, spell_name: str) -> list[AliasSpellbookSpell]:
        needle = str(spell_name).lower()
        return [spell for spell in self.spells if spell.name.lower() == needle]

    def slots_str(self, level: int) -> str:
        slots = self.get_slots(level)
        max_slots = self.get_max_slots(level)
        return f"{slots}/{max_slots}"

    def get_max_slots(self, level: int) -> int:
        slots = self.data.get("max_slots") or {}
        return _safe_int(slots.get(int(level)), 0)

    def get_slots(self, level: int) -> int:
        slots = self.data.get("slots") or {}
        if int(level) == 0:
            return 1
        return _safe_int(slots.get(int(level)), 0)

    def set_slots(self, level: int, value: int, pact: bool = True) -> int:
        slots = self.data.setdefault("slots", {})
        slots[int(level)] = int(value)
        return slots[int(level)]

    def use_slot(self, level: int) -> int:
        current = self.get_slots(level)
        return self.set_slots(level, max(0, current - 1))

    def reset_slots(self) -> None:
        slots = self.data.get("slots") or {}
        max_slots = self.data.get("max_slots") or {}
        for level, maximum in max_slots.items():
            slots[int(level)] = _safe_int(maximum)
        self.data["slots"] = slots

    def reset_pact_slots(self) -> None:
        if self.max_pact_slots is None:
            return
        self.data["num_pact_slots"] = self.max_pact_slots

    def remaining_casts_of(self, spell: str, level: int) -> str:
        return self.slots_str(level)

    def cast(self, spell: str, level: int) -> str:
        self.use_slot(level)
        return f"Casted {spell} at level {level}"

    def can_cast(self, spell: str, level: int) -> bool:
        return self.get_slots(level) > 0

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, str):
            return False
        return any(spell.name.lower() == item.lower() for spell in self.spells)


@dataclass
class AliasCoinpurse(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["pp", "gp", "ep", "sp", "cp", "total"]
    METHODS: ClassVar[list[str]] = ["coin_str", "compact_str", "modify_coins", "set_coins", "autoconvert", "get_coins"]

    def __getattr__(self, item: str) -> Any:
        if item in {"pp", "gp", "ep", "sp", "cp"}:
            return _safe_int(self.data.get(item), 0)
        return self.data.get(item)

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))

    @property
    def total(self) -> float:
        return self.pp * 10 + self.gp + self.ep * 0.5 + self.sp * 0.1 + self.cp * 0.01

    def coin_str(self, cointype: str) -> str:
        cointype = str(cointype)
        value = getattr(self, cointype)
        return f"{value} {cointype}"

    def compact_str(self) -> str:
        return f"{self.total:.2f} gp"

    def modify_coins(
        self,
        pp: int = 0,
        gp: int = 0,
        ep: int = 0,
        sp: int = 0,
        cp: int = 0,
        autoconvert: bool = True,
    ) -> dict[str, Any]:
        self.data["pp"] = self.pp + int(pp)
        self.data["gp"] = self.gp + int(gp)
        self.data["ep"] = self.ep + int(ep)
        self.data["sp"] = self.sp + int(sp)
        self.data["cp"] = self.cp + int(cp)
        return self.get_coins()

    def set_coins(self, pp: int, gp: int, ep: int, sp: int, cp: int) -> None:
        self.data["pp"] = int(pp)
        self.data["gp"] = int(gp)
        self.data["ep"] = int(ep)
        self.data["sp"] = int(sp)
        self.data["cp"] = int(cp)

    def autoconvert(self) -> None:
        total_cp = int(self.total * 100)
        self.data["pp"], remainder = divmod(total_cp, 1000)
        self.data["gp"], remainder = divmod(remainder, 100)
        self.data["ep"], remainder = divmod(remainder, 50)
        self.data["sp"], self.data["cp"] = divmod(remainder, 10)

    def get_coins(self) -> dict[str, Any]:
        return {"pp": self.pp, "gp": self.gp, "ep": self.ep, "sp": self.sp, "cp": self.cp, "total": self.total}


@dataclass
class AliasDeathSaves(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["successes", "fails"]
    METHODS: ClassVar[list[str]] = ["succeed", "fail", "is_stable", "is_dead", "reset"]

    @property
    def successes(self) -> int:
        return _safe_int(self.data.get("successes"), 0)

    @property
    def fails(self) -> int:
        return _safe_int(self.data.get("fails"), 0)

    def succeed(self, num: int = 1) -> None:
        self.data["successes"] = self.successes + int(num)

    def fail(self, num: int = 1) -> None:
        self.data["fails"] = self.fails + int(num)

    def is_stable(self) -> bool:
        return self.successes >= 3 and self.fails < 3

    def is_dead(self) -> bool:
        return self.fails >= 3

    def reset(self) -> None:
        self.data["successes"] = 0
        self.data["fails"] = 0

    def __str__(self) -> str:
        return f"{self.successes} successes / {self.fails} failures"


@dataclass
class AliasAction(_DirMixin):
    data: Mapping[str, Any]
    parent_statblock: Mapping[str, Any]
    ATTRS: ClassVar[list[str]] = ["name", "activation_type", "activation_type_name", "description", "snippet"]
    METHODS: ClassVar[list[str]] = []

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Action"))

    @property
    def activation_type(self) -> int | None:
        raw = self.data.get("activation_type")
        return int(raw) if raw is not None else None

    @property
    def activation_type_name(self) -> str | None:
        val = self.data.get("activation_type_name")
        return str(val) if val is not None else None

    @property
    def description(self) -> str:
        return str(self.data.get("description", ""))

    @property
    def snippet(self) -> str:
        return str(self.data.get("snippet", self.description))

    def __str__(self) -> str:
        return f"**{self.name}**: {self.description}"

    def __getitem__(self, item: str) -> Any:
        if isinstance(item, int) or (isinstance(item, str) and item.isdigit()):
            raise TypeError("AliasAction attributes must be accessed by name (e.g., 'name', 'activation_type').")
        return getattr(self, str(item))

    def __iter__(self):
        return iter(())

    def __len__(self) -> int:
        return 0


@dataclass
class AliasCustomCounter(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = [
        "name",
        "title",
        "desc",
        "value",
        "max",
        "min",
        "reset_on",
        "display_type",
        "reset_to",
        "reset_by",
    ]
    METHODS: ClassVar[list[str]] = ["mod", "set", "reset", "full_str"]

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Counter"))

    @property
    def title(self) -> str | None:
        val = self.data.get("title")
        return str(val) if val is not None else None

    @property
    def desc(self) -> str | None:
        val = self.data.get("desc")
        return str(val) if val is not None else None

    @property
    def value(self) -> int:
        return _safe_int(self.data.get("value"), 0)

    @property
    def max(self) -> int:
        return _safe_int(self.data.get("max"), 2**31 - 1)

    @property
    def min(self) -> int:
        return _safe_int(self.data.get("min"), -(2**31))

    @property
    def reset_on(self) -> str | None:
        val = self.data.get("reset_on")
        return str(val) if val is not None else None

    @property
    def display_type(self) -> str | None:
        val = self.data.get("display_type")
        return str(val) if val is not None else None

    @property
    def reset_to(self) -> int | None:
        raw = self.data.get("reset_to")
        return int(raw) if raw is not None else None

    @property
    def reset_by(self) -> str | None:
        val = self.data.get("reset_by")
        return str(val) if val is not None else None

    def mod(self, value: int, strict: bool = False) -> int:
        return self.set(self.value + int(value), strict)

    def set(self, new_value: int, strict: bool = False) -> int:
        val = int(new_value)
        if strict:
            if val > self.max or val < self.min:
                raise ValueError("Counter out of bounds")
        val = max(self.min, min(val, self.max))
        self.data["value"] = val
        return val

    def reset(self) -> dict[str, Any]:
        if self.reset_to is not None:
            target = self.reset_to
        elif self.reset_by is not None:
            target = self.value + _safe_int(self.reset_by, 0)
        else:
            target = self.max
        old_value = self.value
        new_value = self.set(target)
        return {"new_value": new_value, "old_value": old_value, "target_value": target, "delta": new_value - old_value}

    def full_str(self, include_name: bool = False) -> str:
        prefix = f"**{self.name}**\n" if include_name else ""
        content = f"{self.value}/{self.max}"
        if self.display_type:
            content = self.display_type
        return prefix + content

    def __str__(self) -> str:
        return f"{self.value}/{self.max}"

    def __getitem__(self, item: str) -> Any:
        if isinstance(item, int) or (isinstance(item, str) and item.isdigit()):
            raise TypeError("CustomCounter indices must be strings (e.g., 'value', 'max').")
        return getattr(self, str(item))

    def __iter__(self):
        return iter(())

    def __len__(self) -> int:
        return 0


# === StatBlock + Character ===
@dataclass
class AliasStatBlock(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = [
        "name",
        "stats",
        "levels",
        "attacks",
        "skills",
        "saves",
        "resistances",
        "ac",
        "max_hp",
        "hp",
        "temp_hp",
        "spellbook",
        "creature_type",
    ]
    METHODS: ClassVar[list[str]] = ["set_hp", "modify_hp", "hp_str", "reset_hp", "set_temp_hp"]

    def _prof_bonus(self) -> int:
        stats = self.data.get("stats") or {}
        if "prof_bonus" in stats:
            return _safe_int(stats.get("prof_bonus"), 2)
        levels = self.data.get("levels") or self.data.get("class_levels") or {}
        total_level = sum(_safe_int(val, 0) for val in levels.values()) or 1
        return max(2, 2 + (math.ceil(total_level / 4)))

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Statblock"))

    @property
    def stats(self) -> AliasBaseStats:
        return AliasBaseStats(self.data.get("stats") or {}, prof_bonus_override=self._prof_bonus())

    @property
    def levels(self) -> AliasLevels:
        return AliasLevels(self.data.get("levels") or self.data.get("class_levels") or {})

    @property
    def attacks(self) -> AliasAttackList:
        return AliasAttackList(self.data.get("attacks") or [], self.data)

    @property
    def skills(self) -> AliasSkills:
        abilities = {
            "strength": self.stats.strength,
            "dexterity": self.stats.dexterity,
            "constitution": self.stats.constitution,
            "intelligence": self.stats.intelligence,
            "wisdom": self.stats.wisdom,
            "charisma": self.stats.charisma,
        }
        return AliasSkills(self.data.get("skills") or {}, self._prof_bonus(), abilities)

    @property
    def saves(self) -> AliasSaves:
        abilities = {
            "strength": self.stats.strength,
            "dexterity": self.stats.dexterity,
            "constitution": self.stats.constitution,
            "intelligence": self.stats.intelligence,
            "wisdom": self.stats.wisdom,
            "charisma": self.stats.charisma,
            "str": self.stats.strength,
            "dex": self.stats.dexterity,
            "con": self.stats.constitution,
            "int": self.stats.intelligence,
            "wis": self.stats.wisdom,
            "cha": self.stats.charisma,
        }
        return AliasSaves(self.data.get("saves") or {}, self._prof_bonus(), abilities)

    @property
    def resistances(self) -> AliasResistances:
        return AliasResistances(self.data.get("resistances") or {})

    @property
    def ac(self) -> int | None:
        raw = self.data.get("ac")
        return int(raw) if raw is not None else None

    @property
    def max_hp(self) -> int | None:
        raw = self.data.get("max_hp")
        return int(raw) if raw is not None else None

    @property
    def hp(self) -> int | None:
        raw = self.data.get("hp")
        return int(raw) if raw is not None else None

    @property
    def temp_hp(self) -> int:
        return _safe_int(self.data.get("temp_hp"), 0)

    @property
    def spellbook(self) -> AliasSpellbook:
        return AliasSpellbook(self.data.get("spellbook") or {})

    @property
    def creature_type(self) -> str | None:
        val = self.data.get("creature_type")
        return str(val) if val is not None else None

    def set_hp(self, new_hp: int) -> int:
        self.data["hp"] = int(new_hp)
        return self.data["hp"]

    def modify_hp(self, amount: int, ignore_temp: bool = False, overflow: bool = True) -> int:
        hp = self.hp or 0
        new_hp = hp + int(amount)
        if not overflow and self.max_hp is not None:
            new_hp = max(0, min(new_hp, self.max_hp))
        self.data["hp"] = new_hp
        return new_hp

    def hp_str(self) -> str:
        return f"{self.hp}/{self.max_hp} (+{self.temp_hp} temp)"

    def reset_hp(self) -> int:
        if self.max_hp is not None:
            self.data["hp"] = self.max_hp
        self.data["temp_hp"] = 0
        return self.hp or 0

    def set_temp_hp(self, new_temp: int) -> int:
        self.data["temp_hp"] = int(new_temp)
        return self.temp_hp

    def __getattr__(self, item: str) -> Any:
        return self.data.get(item)

    def __getitem__(self, item: str) -> Any:
        if isinstance(item, int) or (isinstance(item, str) and item.isdigit()):
            raise TypeError("AliasStatBlock attributes must be accessed by name (e.g., 'name', 'stats').")
        return getattr(self, str(item))

    def __iter__(self):
        return iter(())

    def __len__(self) -> int:
        return 0


@dataclass
class CharacterAPI(AliasStatBlock):
    ATTRS: ClassVar[list[str]] = AliasStatBlock.ATTRS + [
        "actions",
        "coinpurse",
        "csettings",
        "race",
        "background",
        "owner",
        "upstream",
        "sheet_type",
        "cvars",
        "consumables",
        "death_saves",
        "description",
        "image",
    ]
    METHODS: ClassVar[list[str]] = AliasStatBlock.METHODS + [
        "cc",
        "get_cc",
        "get_cc_max",
        "get_cc_min",
        "set_cc",
        "mod_cc",
        "delete_cc",
        "create_cc_nx",
        "create_cc",
        "edit_cc",
        "cc_exists",
        "cc_str",
        "get_cvar",
        "set_cvar",
        "set_cvar_nx",
        "delete_cvar",
    ]

    def _consumable_map(self) -> MutableMapping[str, MutableMapping[str, Any]]:
        consumables = self.data.setdefault("consumables", {})
        if isinstance(consumables, list):
            # normalize list payloads to map
            mapped: dict[str, MutableMapping[str, Any]] = {}
            for item in consumables:
                name = str(item.get("name", f"cc_{len(mapped)}"))
                mapped[name] = dict(item)
            self.data["consumables"] = mapped
            consumables = mapped
        return consumables  # type: ignore[return-value]

    @property
    def actions(self) -> list[AliasAction]:
        acts = self.data.get("actions") or []
        return [AliasAction(a, self.data) for a in acts]

    @property
    def coinpurse(self) -> AliasCoinpurse:
        return AliasCoinpurse(self.data.get("coinpurse") or {})

    @property
    def csettings(self) -> Mapping[str, Any]:
        return self.data.get("csettings", {})

    @property
    def race(self) -> str | None:
        val = self.data.get("race")
        return str(val) if val is not None else None

    @property
    def background(self) -> str | None:
        val = self.data.get("background")
        return str(val) if val is not None else None

    @property
    def owner(self) -> int | None:
        raw = self.data.get("owner")
        return int(raw) if raw is not None else None

    @property
    def upstream(self) -> str | None:
        val = self.data.get("upstream")
        return str(val) if val is not None else None

    @property
    def sheet_type(self) -> str | None:
        val = self.data.get("sheet_type")
        return str(val) if val is not None else None

    @property
    def cvars(self) -> Mapping[str, Any]:
        return dict(self.data.get("cvars") or {})

    def get_cvar(self, name: str, default: Any = None) -> Any:
        return self.data.setdefault("cvars", {}).get(str(name), default)

    def set_cvar(self, name: str, val: Any) -> Any:
        self.data.setdefault("cvars", {})[str(name)] = val
        return val

    def set_cvar_nx(self, name: str, val: Any) -> Any:
        cvars = self.data.setdefault("cvars", {})
        return cvars.setdefault(str(name), val)

    def delete_cvar(self, name: str) -> Any:
        return self.data.setdefault("cvars", {}).pop(str(name), None)

    @property
    def consumables(self) -> list[AliasCustomCounter]:
        return [AliasCustomCounter(v) for v in self._consumable_map().values()]

    def cc(self, name: str) -> AliasCustomCounter:
        return AliasCustomCounter(self._consumable_map()[str(name)])

    def get_cc(self, name: str) -> int:
        return self.cc(name).value

    def get_cc_max(self, name: str) -> int:
        return self.cc(name).max

    def get_cc_min(self, name: str) -> int:
        return self.cc(name).min

    def set_cc(self, name: str, value: int | None = None, maximum: int | None = None, minimum: int | None = None) -> int:
        con = self._consumable_map().setdefault(str(name), {"name": str(name)})
        if value is not None:
            con["value"] = int(value)
        if maximum is not None:
            con["max"] = int(maximum)
        if minimum is not None:
            con["min"] = int(minimum)
        return _safe_int(con.get("value"), 0)

    def mod_cc(self, name: str, val: int, strict: bool = False) -> int:
        counter = self.cc(name)
        return counter.mod(val, strict)

    def delete_cc(self, name: str) -> Any:
        return self._consumable_map().pop(str(name), None)

    def create_cc_nx(
        self,
        name: str,
        minVal: str | None = None,
        maxVal: str | None = None,
        reset: str | None = None,
        dispType: str | None = None,
        reset_to: str | None = None,
        reset_by: str | None = None,
        title: str | None = None,
        desc: str | None = None,
        initial_value: str | None = None,
    ) -> AliasCustomCounter:
        if not self.cc_exists(name):
            self.create_cc(
                name,
                minVal=minVal,
                maxVal=maxVal,
                reset=reset,
                dispType=dispType,
                reset_to=reset_to,
                reset_by=reset_by,
                title=title,
                desc=desc,
                initial_value=initial_value,
            )
        return self.cc(name)

    def create_cc(
        self,
        name: str,
        minVal: str | None = None,
        maxVal: str | None = None,
        reset: str | None = None,
        dispType: str | None = None,
        reset_to: str | None = None,
        reset_by: str | None = None,
        title: str | None = None,
        desc: str | None = None,
        initial_value: str | None = None,
    ) -> AliasCustomCounter:
        payload = {
            "name": str(name),
            "min": _safe_int(minVal, -(2**31)) if minVal is not None else -(2**31),
            "max": _safe_int(maxVal, 2**31 - 1) if maxVal is not None else 2**31 - 1,
            "reset_on": reset,
            "display_type": dispType,
            "reset_to": _safe_int(reset_to) if reset_to is not None else None,
            "reset_by": reset_by,
            "title": title,
            "desc": desc,
            "value": _safe_int(initial_value, 0) if initial_value is not None else 0,
        }
        self._consumable_map()[str(name)] = payload
        return AliasCustomCounter(payload)

    def edit_cc(
        self,
        name: str,
        minVal: Any = UNSET,
        maxVal: Any = UNSET,
        reset: Any = UNSET,
        dispType: Any = UNSET,
        reset_to: Any = UNSET,
        reset_by: Any = UNSET,
        title: Any = UNSET,
        desc: Any = UNSET,
        new_name: str | None = None,
    ) -> AliasCustomCounter:
        counter = dict(self._consumable_map().get(str(name)) or {"name": str(name)})
        for key, val in (
            ("min", minVal),
            ("max", maxVal),
            ("reset_on", reset),
            ("display_type", dispType),
            ("reset_to", reset_to),
            ("reset_by", reset_by),
            ("title", title),
            ("desc", desc),
        ):
            if val is not UNSET:
                counter[key] = val
        counter["name"] = str(new_name) if new_name else counter.get("name", str(name))
        self._consumable_map().pop(str(name), None)
        self._consumable_map()[counter["name"]] = counter
        return AliasCustomCounter(counter)

    def cc_exists(self, name: str) -> bool:
        return str(name) in self._consumable_map()

    def cc_str(self, name: str) -> str:
        return str(self.cc(name))

    @property
    def death_saves(self) -> AliasDeathSaves:
        return AliasDeathSaves(self.data.get("death_saves") or {})

    @property
    def description(self) -> str | None:
        val = self.data.get("description")
        return str(val) if val is not None else None

    @property
    def image(self) -> str | None:
        val = self.data.get("image")
        return str(val) if val is not None else None


# === Combat API ===
MAX_COMBAT_METADATA_SIZE = 100000


@dataclass
class SimpleEffect(_DirMixin):
    data: MutableMapping[str, Any]
    ATTRS: ClassVar[list[str]] = [
        "name",
        "duration",
        "remaining",
        "effect",
        "attacks",
        "buttons",
        "conc",
        "desc",
        "ticks_on_end",
        "combatant_name",
        "parent",
        "children",
    ]
    METHODS: ClassVar[list[str]] = ["set_parent"]

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Effect"))

    @property
    def duration(self) -> int | None:
        raw = self.data.get("duration")
        return int(raw) if raw is not None else None

    @property
    def remaining(self) -> int | None:
        raw = self.data.get("remaining")
        return int(raw) if raw is not None else None

    @property
    def effect(self) -> Mapping[str, Any]:
        return self.data.get("effects") or self.data.get("effect") or {}

    @property
    def attacks(self) -> list[Mapping[str, Any]]:
        return list(self.data.get("attacks") or [])

    @property
    def buttons(self) -> list[Mapping[str, Any]]:
        return list(self.data.get("buttons") or [])

    @property
    def conc(self) -> bool:
        return bool(self.data.get("conc") or self.data.get("concentration", False))

    @property
    def desc(self) -> str | None:
        val = self.data.get("desc")
        return str(val) if val is not None else None

    @property
    def ticks_on_end(self) -> bool:
        return bool(self.data.get("ticks_on_end") or self.data.get("end_on_turn_end", False))

    @property
    def combatant_name(self) -> str | None:
        val = self.data.get("combatant_name")
        return str(val) if val is not None else None

    @property
    def parent(self) -> "SimpleEffect" | None:
        parent = self.data.get("parent")
        return SimpleEffect(parent) if parent else None

    @property
    def children(self) -> list["SimpleEffect"]:
        return [SimpleEffect(c) for c in self.data.get("children", [])]

    def set_parent(self, parent: "SimpleEffect") -> None:
        if not isinstance(parent, SimpleEffect):
            raise TypeError("Parent effect must be a SimpleEffect.")
        self.data["parent"] = parent.data

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class SimpleCombatant(AliasStatBlock):
    ATTRS: ClassVar[list[str]] = AliasStatBlock.ATTRS + [
        "effects",
        "init",
        "initmod",
        "type",
        "note",
        "controller",
        "group",
        "race",
        "monster_name",
        "is_hidden",
        "id",
    ]
    METHODS: ClassVar[list[str]] = AliasStatBlock.METHODS + [
        "save",
        "damage",
        "set_ac",
        "set_maxhp",
        "set_init",
        "set_name",
        "set_group",
        "set_note",
        "get_effect",
        "add_effect",
        "remove_effect",
    ]

    def __post_init__(self) -> None:
        super().__post_init__()
        self.data.setdefault("type", "combatant")

    @property
    def id(self) -> str | None:
        val = self.data.get("id")
        return str(val) if val is not None else None

    @property
    def effects(self) -> list[SimpleEffect]:
        return [SimpleEffect(e) for e in self.data.get("effects", [])]

    @property
    def init(self) -> int:
        return _safe_int(self.data.get("init"), 0)

    @property
    def initmod(self) -> int:
        return _safe_int(self.data.get("initmod"), 0)

    @property
    def type(self) -> str:
        return str(self.data.get("type", "combatant"))

    @property
    def note(self) -> str | None:
        val = self.data.get("note")
        return str(val) if val is not None else None

    @property
    def controller(self) -> int | None:
        raw = self.data.get("controller")
        return int(raw) if raw is not None else None

    @property
    def group(self) -> str | None:
        val = self.data.get("group")
        return str(val) if val is not None else None

    @property
    def race(self) -> str | None:
        val = self.data.get("race")
        return str(val) if val is not None else None

    @property
    def monster_name(self) -> str | None:
        val = self.data.get("monster_name")
        return str(val) if val is not None else None

    @property
    def is_hidden(self) -> bool:
        return bool(self.data.get("is_hidden", False))

    def save(self, ability: str, adv: bool | None = None) -> SimpleRollResult:
        roll_expr = self.saves.get(ability).d20(base_adv=adv)
        try:
            roll_result = d20.roll(roll_expr)
        except Exception:
            roll_result = d20.roll("0")
        return SimpleRollResult(roll_result)

    def damage(
        self,
        dice_str: str,
        crit: bool = False,
        d=None,
        c=None,
        critdice: int = 0,
        overheal: bool = False,
    ) -> dict[str, Any]:
        expr = str(dice_str)
        if crit:
            expr = f"({expr})*2"
        if d is not None:
            expr = f"({expr})+({d})"
        if c is not None and crit:
            expr = f"({expr})+({c})"
        if critdice:
            expr = f"({expr})+{int(critdice)}"
        try:
            roll_result = d20.roll(expr)
        except Exception:
            roll_result = d20.roll("0")
        label = "Damage (CRIT!)" if crit else "Damage"
        return {"damage": f"**{label}**: {roll_result}", "total": roll_result.total, "roll": SimpleRollResult(roll_result)}

    def set_ac(self, ac: int) -> None:
        self.data["ac"] = int(ac)

    def set_maxhp(self, maxhp: int) -> None:
        self.data["max_hp"] = int(maxhp)

    def set_init(self, init: int) -> None:
        self.data["init"] = int(init)

    def set_name(self, name: str) -> None:
        self.data["name"] = str(name)

    def set_group(self, group: str | None) -> str | None:
        self.data["group"] = str(group) if group is not None else None
        return self.group

    def set_note(self, note: str) -> None:
        self.data["note"] = str(note) if note is not None else None

    def get_effect(self, name: str, strict: bool = False) -> SimpleEffect | None:
        name_lower = str(name).lower()
        for effect in self.effects:
            if strict and effect.name.lower() == name_lower:
                return effect
            if not strict and name_lower in effect.name.lower():
                return effect
        return None

    def add_effect(
        self,
        name: str,
        args: str | None = None,
        duration: int | None = None,
        concentration: bool = False,
        parent: SimpleEffect | None = None,
        end: bool = False,
        desc: str | None = None,
        passive_effects: dict | None = None,
        attacks: list[dict] | None = None,
        buttons: list[dict] | None = None,
        tick_on_combatant_id: str | None = None,
    ) -> SimpleEffect:
        duration_val = int(duration) if duration is not None else None
        desc_val = str(desc) if desc is not None else None
        payload: dict[str, Any] = {
            "name": str(name),
            "duration": duration_val,
            "remaining": duration_val,
            "args": str(args) if args is not None else None,
            "desc": desc_val,
            "concentration": bool(concentration),
            "conc": bool(concentration),
            "ticks_on_end": end,
            "effect": dict(passive_effects or {}),
            "attacks": list(attacks or []),
            "buttons": list(buttons or []),
            "combatant_name": self.name,
        }
        if tick_on_combatant_id is not None:
            payload["tick_on_combatant_id"] = str(tick_on_combatant_id)
        if parent is not None:
            if not isinstance(parent, SimpleEffect):
                raise TypeError("Parent effect must be a SimpleEffect.")
            payload["parent"] = parent.data
        effects = self.data.setdefault("effects", [])
        existing = self.get_effect(name, strict=True)
        if existing:
            try:
                effects.remove(existing.data)
            except ValueError:
                pass
        effects.append(payload)
        return SimpleEffect(payload)

    def remove_effect(self, name: str, strict: bool = False) -> None:
        effect = self.get_effect(name, strict)
        if effect:
            try:
                self.data.setdefault("effects", []).remove(effect.data)
            except ValueError:
                pass


@dataclass
class SimpleGroup(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["combatants", "type", "init", "name", "id"]
    METHODS: ClassVar[list[str]] = ["get_combatant", "set_init"]

    def __post_init__(self) -> None:
        super().__post_init__()
        self.data.setdefault("type", "group")

    @property
    def combatants(self) -> list[SimpleCombatant]:
        return [SimpleCombatant(c) for c in self.data.get("combatants", [])]

    @property
    def type(self) -> str:
        return str(self.data.get("type", "group"))

    @property
    def init(self) -> int:
        return _safe_int(self.data.get("init"), 0)

    @property
    def name(self) -> str:
        return str(self.data.get("name", "Group"))

    @property
    def id(self) -> str | None:
        val = self.data.get("id")
        return str(val) if val is not None else None

    def get_combatant(self, name: str, strict: bool | None = None) -> SimpleCombatant | None:
        name_lower = str(name).lower()
        for combatant in self.combatants:
            if strict is True and combatant.name.lower() == name_lower:
                return combatant
            if strict is None and combatant.name.lower() == name_lower:
                return combatant
            if strict is False and name_lower in combatant.name.lower():
                return combatant
        if strict is None:
            for combatant in self.combatants:
                if name_lower in combatant.name.lower():
                    return combatant
        return None

    def set_init(self, init: int) -> None:
        self.data["init"] = int(init)

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


@dataclass
class SimpleCombat(_DirMixin):
    data: MutableMapping[str, Any] = field(default_factory=dict)
    ATTRS: ClassVar[list[str]] = ["combatants", "groups", "me", "current", "name", "round_num", "turn_num", "metadata"]
    METHODS: ClassVar[list[str]] = ["get_combatant", "get_group", "set_metadata", "get_metadata", "delete_metadata", "set_round", "end_round"]

    @property
    def combatants(self) -> list[SimpleCombatant]:
        return [SimpleCombatant(c) for c in self.data.get("combatants", [])]

    @property
    def groups(self) -> list[SimpleGroup]:
        return [SimpleGroup(g) for g in self.data.get("groups", [])]

    @property
    def me(self) -> SimpleCombatant | None:
        me_data = self.data.get("me")
        return SimpleCombatant(me_data) if me_data is not None else None

    @property
    def current(self) -> SimpleCombatant | SimpleGroup | None:
        cur = self.data.get("current")
        if cur is None:
            return None
        if cur.get("type") == "group":
            return SimpleGroup(cur)
        return SimpleCombatant(cur)

    @property
    def name(self) -> str | None:
        val = self.data.get("name")
        return str(val) if val is not None else None

    @property
    def round_num(self) -> int:
        return _safe_int(self.data.get("round_num"), 1)

    @property
    def turn_num(self) -> int:
        return _safe_int(self.data.get("turn_num"), 1)

    @property
    def metadata(self) -> MutableMapping[str, Any]:
        return self.data.setdefault("metadata", {})

    def get_combatant(self, name: str, strict: bool | None = None) -> SimpleCombatant | None:
        name_lower = str(name).lower()
        for combatant in self.combatants:
            if strict is True and combatant.name.lower() == name_lower:
                return combatant
            if strict is None and combatant.name.lower() == name_lower:
                return combatant
            if strict is False and name_lower in combatant.name.lower():
                return combatant
        if strict is None:
            for combatant in self.combatants:
                if name_lower in combatant.name.lower():
                    return combatant
        return None

    def get_group(self, name: str, strict: bool | None = None) -> SimpleGroup | None:
        name_lower = str(name).lower()
        for group in self.groups:
            if strict is True and group.name.lower() == name_lower:
                return group
            if strict is None and group.name.lower() == name_lower:
                return group
            if strict is False and name_lower in group.name.lower():
                return group
        if strict is None:
            for group in self.groups:
                if name_lower in group.name.lower():
                    return group
        return None

    def set_metadata(self, k: str, v: str) -> None:
        key = str(k)
        value = str(v)
        existing = {str(ke): str(va) for ke, va in self.metadata.items() if str(ke) != key}
        proposed_size = sum(len(ke) + len(va) for ke, va in existing.items()) + len(key) + len(value)
        if proposed_size > MAX_COMBAT_METADATA_SIZE:
            raise ValueError("Combat metadata is too large")
        self.metadata[key] = value

    def get_metadata(self, k: str, default: Any = None) -> Any:
        return self.metadata.get(str(k), default)

    def delete_metadata(self, k: str) -> Any:
        return self.metadata.pop(str(k), None)

    def set_round(self, round_num: int) -> None:
        self.data["round_num"] = int(round_num)

    def end_round(self) -> None:
        self.data["turn_num"] = 0
        self.data["round_num"] = self.round_num + 1

    def __getattr__(self, item: str) -> Any:
        return self.data.get(item)

    def __getitem__(self, item: str) -> Any:
        return getattr(self, str(item))


# Backwards-compatible aliases for existing consumers
CombatantAPI = SimpleCombatant
SimpleEffectAPI = SimpleEffect
SimpleGroupAPI = SimpleGroup
CombatAPI = SimpleCombat



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict

import draconic
from lsprotocol import types
from pygls import uris
from pygls.lsp.server import LanguageServer

from .config import AvraeLSConfig, load_config
from .context import ContextBuilder
from .diagnostics import DiagnosticProvider
from .runtime import MockExecutor
from .alias_preview import render_alias_command, simulate_command
from .parser import find_draconic_blocks
from .signature_help import load_signatures, signature_help_for_code
from .completions import gather_suggestions, completion_items_for_position, hover_for_position
from .symbols import build_symbol_table, document_symbols, find_definition_range
from .argument_parsing import apply_argument_parsing

__version__ = "0.1.0"
log = logging.getLogger(__name__)

RUN_ALIAS_COMMAND = "avrae.runAlias"
REFRESH_CONFIG_COMMAND = "avrae.reloadConfig"
REFRESH_GVARS_COMMAND = "avrae.refreshGvars"
LEVEL_TO_SEVERITY = {
    "error": types.DiagnosticSeverity.Error,
    "warning": types.DiagnosticSeverity.Warning,
    "info": types.DiagnosticSeverity.Information,
}


@dataclass
class ServerState:
    config: AvraeLSConfig
    context_builder: ContextBuilder
    diagnostics: DiagnosticProvider
    executor: MockExecutor
    warnings: list[str] = field(default_factory=list)


class AvraeLanguageServer(LanguageServer):
    def __init__(self):
        super().__init__(
            name="avrae-ls",
            version=__version__,
            text_document_sync_kind=types.TextDocumentSyncKind.Incremental,
        )
        self._state: ServerState | None = None
        self._workspace_root: Path | None = None
        self._signatures: Dict[str, Any] = load_signatures()

    @property
    def state(self) -> ServerState:
        if self._state is None:
            raise RuntimeError("Server has not been initialized")
        return self._state

    @property
    def workspace_root(self) -> Path:
        if self._workspace_root is None:
            return Path.cwd()
        return self._workspace_root

    def load_workspace(self, root: Path) -> None:
        config, warnings = load_config(root)
        executor = MockExecutor()
        context_builder = ContextBuilder(config)
        diagnostics = DiagnosticProvider(executor, config.diagnostics)
        self._state = ServerState(
            config=config,
            context_builder=context_builder,
            diagnostics=diagnostics,
            executor=executor,
            warnings=list(warnings),
        )
        self._workspace_root = root
        log.info("Loaded workspace at %s", root)


ls = AvraeLanguageServer()


@ls.feature(types.INITIALIZE)
def on_initialize(server: AvraeLanguageServer, params: types.InitializeParams):
    root_uri = params.root_uri or (params.workspace_folders[0].uri if params.workspace_folders else None)
    root_path = Path(uris.to_fs_path(root_uri)) if root_uri else Path.cwd()
    server.load_workspace(root_path)


@ls.feature(types.INITIALIZED)
async def on_initialized(server: AvraeLanguageServer, params: types.InitializedParams):
    for warning in server.state.warnings:
        server.window_log_message(
            types.LogMessageParams(type=types.MessageType.Warning, message=warning)
        )


@ls.feature(types.TEXT_DOCUMENT_DID_OPEN)
async def did_open(server: AvraeLanguageServer, params: types.DidOpenTextDocumentParams):
    await _publish_diagnostics(server, params.text_document.uri)


@ls.feature(types.TEXT_DOCUMENT_DID_CHANGE)
async def did_change(server: AvraeLanguageServer, params: types.DidChangeTextDocumentParams):
    await _publish_diagnostics(server, params.text_document.uri)


@ls.feature(types.TEXT_DOCUMENT_DID_SAVE)
async def did_save(server: AvraeLanguageServer, params: types.DidSaveTextDocumentParams):
    await _publish_diagnostics(server, params.text_document.uri)


@ls.feature(types.WORKSPACE_DID_CHANGE_CONFIGURATION)
async def did_change_config(server: AvraeLanguageServer, params: types.DidChangeConfigurationParams):
    server.load_workspace(server.workspace_root)
    for warning in server.state.warnings:
        server.window_log_message(
            types.LogMessageParams(type=types.MessageType.Warning, message=warning)
        )


@ls.feature(types.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def on_document_symbol(server: AvraeLanguageServer, params: types.DocumentSymbolParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    symbols = document_symbols(doc.source)
    return symbols


@ls.feature(types.TEXT_DOCUMENT_DEFINITION)
def on_definition(server: AvraeLanguageServer, params: types.DefinitionParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    table = build_symbol_table(doc.source)
    word = doc.word_at_position(params.position)
    rng = find_definition_range(table, word)
    if rng is None:
        return None
    return types.Location(uri=params.text_document.uri, range=rng)


@ls.feature(types.WORKSPACE_SYMBOL)
def on_workspace_symbol(server: AvraeLanguageServer, params: types.WorkspaceSymbolParams):
    symbols: list[types.SymbolInformation] = []
    query = (params.query or "").lower()
    for uri, doc in server.workspace.text_documents.items():
        table = build_symbol_table(doc.source)
        for entry in table.entries:
            if query and query not in entry.name.lower():
                continue
            symbols.append(
                types.SymbolInformation(
                    name=entry.name,
                    kind=entry.kind,
                    location=types.Location(uri=uri, range=entry.range),
                )
            )
    return symbols


@ls.feature(types.TEXT_DOCUMENT_SIGNATURE_HELP)
def on_signature_help(server: AvraeLanguageServer, params: types.SignatureHelpParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    source = apply_argument_parsing(doc.source)
    blocks = find_draconic_blocks(source)
    pos = params.position
    if not blocks:
        return signature_help_for_code(source, pos.line, pos.character, server._signatures)

    for block in blocks:
        start = block.line_offset
        end = block.line_offset + block.line_count
        if start <= pos.line <= end:
            rel_line = pos.line - start
            help_ = signature_help_for_code(block.code, rel_line, pos.character, server._signatures)
            if help_:
                return help_
    return None


@ls.feature(
    types.TEXT_DOCUMENT_COMPLETION,
    types.CompletionOptions(trigger_characters=["."]),
)
def on_completion(server: AvraeLanguageServer, params: types.CompletionParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    ctx_data = server.state.context_builder.build()
    suggestions = gather_suggestions(ctx_data, server.state.context_builder.gvar_resolver, server._signatures)
    source = apply_argument_parsing(doc.source)
    blocks = find_draconic_blocks(source)
    pos = params.position
    if not blocks:
        return completion_items_for_position(source, pos.line, pos.character, suggestions)

    for block in blocks:
        start = block.line_offset
        end = block.line_offset + block.line_count
        if start <= pos.line <= end:
            rel_line = pos.line - start
            return completion_items_for_position(block.code, rel_line, pos.character, suggestions)
    return []


@ls.feature(types.TEXT_DOCUMENT_HOVER)
def on_hover(server: AvraeLanguageServer, params: types.HoverParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    ctx_data = server.state.context_builder.build()
    pos = params.position
    source = apply_argument_parsing(doc.source)
    blocks = find_draconic_blocks(source)
    if not blocks:
        return hover_for_position(source, pos.line, pos.character, server._signatures, ctx_data, server.state.context_builder.gvar_resolver)

    for block in blocks:
        start = block.line_offset
        end = block.line_offset + block.line_count
        if start <= pos.line <= end:
            rel_line = pos.line - start
            return hover_for_position(block.code, rel_line, pos.character, server._signatures, ctx_data, server.state.context_builder.gvar_resolver)
    return None


@ls.command(RUN_ALIAS_COMMAND)
async def run_alias(server: AvraeLanguageServer, *args: Any):
    payload = args[0] if args else {}
    uri = None
    text = None
    profile = None
    alias_args: list[str] | None = None
    if isinstance(payload, dict):
        uri = payload.get("uri")
        text = payload.get("text")
        profile = payload.get("profile")
        if isinstance(payload.get("args"), list):
            alias_args = [str(a) for a in payload["args"]]

    if text is None and uri:
        doc = server.workspace.get_text_document(uri)
        text = doc.source

    if text is None:
        return {"error": "No alias content supplied"}

    ctx_data = server.state.context_builder.build(profile)
    rendered = await render_alias_command(
        text,
        server.state.executor,
        ctx_data,
        server.state.context_builder.gvar_resolver,
        args=alias_args,
    )
    preview_output, command_name, validation_error = simulate_command(rendered.command)

    response: dict[str, Any] = {
        "stdout": rendered.stdout,
        "result": preview_output if preview_output is not None else rendered.last_value,
        "command": rendered.command,
        "commandName": command_name,
    }
    if rendered.error:
        response["error"] = _format_runtime_error(rendered.error)
    if validation_error:
        response["validationError"] = validation_error
    if uri:
        extra = []
        if rendered.error:
            extra.append(_runtime_diagnostic(rendered.error, server.state.config.diagnostics.runtime_level))
        await _publish_diagnostics(server, uri, profile=profile, extra=extra)
    return response


@ls.command(REFRESH_GVARS_COMMAND)
async def refresh_gvars(server: AvraeLanguageServer, *args: Any):
    payload = args[0] if args else {}
    profile = None
    keys: list[str] | None = None
    if isinstance(payload, dict):
        profile = payload.get("profile")
        raw_keys = payload.get("keys")
        if isinstance(raw_keys, list):
            keys = [str(k) for k in raw_keys]

    ctx_data = server.state.context_builder.build(profile)
    resolver = server.state.context_builder.gvar_resolver
    snapshot = await resolver.refresh(ctx_data.vars.gvars, keys)
    return {"count": len(snapshot), "gvars": snapshot}


@ls.command(REFRESH_CONFIG_COMMAND)
def reload_config(server: AvraeLanguageServer, *args: Any):
    server.load_workspace(server.workspace_root)
    return {"status": "ok"}


async def _publish_diagnostics(
    server: AvraeLanguageServer,
    uri: str,
    profile: str | None = None,
    extra: list[types.Diagnostic] | None = None,
) -> None:
    doc = server.workspace.get_text_document(uri)
    ctx_data = server.state.context_builder.build(profile)
    diags = await server.state.diagnostics.analyze(
        doc.source, ctx_data, server.state.context_builder.gvar_resolver
    )
    if extra:
        diags.extend(extra)
    server.text_document_publish_diagnostics(
        types.PublishDiagnosticsParams(uri=uri, diagnostics=diags, version=doc.version)
    )


def _format_runtime_error(error: BaseException) -> str:
    if isinstance(error, draconic.DraconicException):
        return error.msg
    return str(error)


def _runtime_diagnostic(error: BaseException, level: str) -> types.Diagnostic:
    severity = LEVEL_TO_SEVERITY.get(level.lower(), types.DiagnosticSeverity.Error)
    if isinstance(error, draconic.DraconicSyntaxError):
        rng = types.Range(
            start=types.Position(line=max((error.lineno or 1) - 1, 0), character=max((error.offset or 1) - 1, 0)),
            end=types.Position(
                line=max(((error.end_lineno or error.lineno or 1) - 1), 0),
                character=max(((error.end_offset or error.offset or 1) - 1), 0),
            ),
        )
        message = error.msg
    elif hasattr(error, "node"):
        node = getattr(error, "node")
        rng = types.Range(
            start=types.Position(line=max(getattr(node, "lineno", 1) - 1, 0), character=max(getattr(node, "col_offset", 0), 0)),
            end=types.Position(
                line=max(getattr(node, "end_lineno", getattr(node, "lineno", 1)) - 1, 0),
                character=max(getattr(node, "end_col_offset", getattr(node, "col_offset", 0) + 1), 0),
            ),
        )
        message = getattr(error, "msg", str(error))
    else:
        rng = types.Range(
            start=types.Position(line=0, character=0),
            end=types.Position(line=0, character=1),
        )
        message = str(error)
    return types.Diagnostic(message=message, range=rng, severity=severity, source="avrae-ls-runtime")


def create_server() -> AvraeLanguageServer:
    return ls



import pytest

from avrae_ls.api import SimpleRollResult
from avrae_ls.config import VarSources
from avrae_ls.context import ContextData
from avrae_ls.runtime import MockExecutor


def _ctx():
    return ContextData(vars=VarSources())


@pytest.mark.asyncio
async def test_roll_returns_total():
    executor = MockExecutor()
    result = await executor.run("roll('1d1')", _ctx())
    assert result.error is None
    assert result.value == 1


@pytest.mark.asyncio
async def test_vroll_returns_roll_result():
    executor = MockExecutor()
    result = await executor.run("vroll('1d1')", _ctx())
    assert result.error is None
    assert isinstance(result.value, SimpleRollResult)
    assert result.value.total == 1
    assert result.value.dice
    assert result.value.consolidated()


@pytest.mark.asyncio
async def test_vroll_handles_bad_input():
    executor = MockExecutor()
    result = await executor.run("vroll('not dice')", _ctx())
    assert result.error is None
    assert result.value is None



from __future__ import annotations

import ast
import re
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional

from lsprotocol import types

from .context import ContextData, GVarResolver
from .runtime import _default_builtins
from .api import (
    AliasAction,
    AliasBaseStats,
    AliasCoinpurse,
    AliasContextAPI,
    AliasCustomCounter,
    AliasDeathSaves,
    AliasResistances,
    AliasSaves,
    AliasSkill,
    AliasSkills,
    AliasSpellbook,
    AliasSpellbookSpell,
    AliasAttack,
    AliasAttackList,
    AliasLevels,
    CategoryAPI,
    ChannelAPI,
    CharacterAPI,
    SimpleCombat,
    SimpleCombatant,
    GuildAPI,
    RoleAPI,
    AuthorAPI,
    SimpleEffect,
    SimpleGroup,
    SimpleRollResult,
)
from .signature_help import FunctionSig

TYPE_MAP: Dict[str, object] = {
    "character": CharacterAPI,
    "combat": SimpleCombat,
    "SimpleCombat": SimpleCombat,
    "ctx": AliasContextAPI,
    "SimpleRollResult": SimpleRollResult,
    "stats": AliasBaseStats,
    "levels": AliasLevels,
    "attacks": AliasAttackList,
    "attack": AliasAttack,
    "skills": AliasSkills,
    "skill": AliasSkill,
    "saves": AliasSaves,
    "resistances": AliasResistances,
    "coinpurse": AliasCoinpurse,
    "custom_counter": AliasCustomCounter,
    "consumable": AliasCustomCounter,
    "death_saves": AliasDeathSaves,
    "action": AliasAction,
    "spellbook": AliasSpellbook,
    "spell": AliasSpellbookSpell,
    "guild": GuildAPI,
    "channel": ChannelAPI,
    "category": CategoryAPI,
    "author": AuthorAPI,
    "role": RoleAPI,
    "combatant": SimpleCombatant,
    "SimpleCombatant": SimpleCombatant,
    "group": SimpleGroup,
    "SimpleGroup": SimpleGroup,
    "effect": SimpleEffect,
    "SimpleEffect": SimpleEffect,
}


IDENT_RE = re.compile(r"[A-Za-z_]\w*$")
ATTR_RE = re.compile(r"([A-Za-z_][\w\.]*)\.([A-Za-z_]\w*)?\s*$")
ATTR_AT_CURSOR_RE = re.compile(r"([A-Za-z_][\w\.]*?(?:\(\))?)\.([A-Za-z_]\w*)")


@dataclass
class Suggestion:
    name: str
    kind: types.CompletionItemKind
    detail: str = ""
    documentation: str = ""


def gather_suggestions(
    ctx_data: ContextData,
    resolver: GVarResolver,
    sigs: Dict[str, FunctionSig],
) -> List[Suggestion]:
    suggestions: list[Suggestion] = []

    for name, sig in sigs.items():
        suggestions.append(
            Suggestion(
                name=name,
                kind=types.CompletionItemKind.Function,
                detail=sig.label,
                documentation=sig.doc,
            )
        )

    vars_map = ctx_data.vars.to_initial_names()
    for name in vars_map:
        suggestions.append(Suggestion(name=name, kind=types.CompletionItemKind.Variable, detail="var"))

    gvars = resolver.snapshot()
    for name in gvars:
        suggestions.append(Suggestion(name=name, kind=types.CompletionItemKind.Variable, detail="gvar"))

    for name in _default_builtins().keys():
        if name not in sigs:
            suggestions.append(Suggestion(name=name, kind=types.CompletionItemKind.Function))

    # context helpers
    suggestions.append(Suggestion(name="character", kind=types.CompletionItemKind.Function, detail="Alias character()"))
    suggestions.append(Suggestion(name="combat", kind=types.CompletionItemKind.Function, detail="Alias combat()"))
    suggestions.append(Suggestion(name="ctx", kind=types.CompletionItemKind.Variable, detail="Alias context"))

    return suggestions


def completion_items_for_position(
    code: str,
    line: int,
    character: int,
    suggestions: Iterable[Suggestion],
) -> List[types.CompletionItem]:
    line_text = _line_text_to_cursor(code, line, character)
    attr_match = ATTR_RE.search(line_text)
    if attr_match:
        receiver = attr_match.group(1)
        attr_prefix = attr_match.group(2) or ""
        sanitized = _sanitize_incomplete_line(code, line, character)
        type_map = _infer_type_map(sanitized)
        return _attribute_completions(receiver, attr_prefix, sanitized, type_map)

    prefix = _current_prefix(line_text)
    items: list[types.CompletionItem] = []
    for sugg in suggestions:
        if prefix and not sugg.name.startswith(prefix):
            continue
        items.append(
            types.CompletionItem(
                label=sugg.name,
                kind=sugg.kind,
                detail=sugg.detail or None,
                documentation=sugg.documentation or None,
            )
        )
    return items


def _attribute_completions(receiver: str, prefix: str, code: str, type_map: Dict[str, str] | None = None) -> List[types.CompletionItem]:
    items: list[types.CompletionItem] = []
    type_key = _resolve_type_name(receiver, code, type_map)
    attrs, methods = _type_meta(type_key)
    detail = f"{type_key}()"

    for name in attrs:
        if prefix and not name.startswith(prefix):
            continue
        items.append(
            types.CompletionItem(
                label=name,
                kind=types.CompletionItemKind.Field,
                detail=detail,
            )
        )
    for name in methods:
        if prefix and not name.startswith(prefix):
            continue
        items.append(
            types.CompletionItem(
                label=name,
                kind=types.CompletionItemKind.Method,
                detail=f"{detail} method",
            )
        )
    return items


def hover_for_position(
    code: str,
    line: int,
    character: int,
    sigs: Dict[str, FunctionSig],
    ctx_data: ContextData,
    resolver: GVarResolver,
) -> Optional[types.Hover]:
    line_text = _line_text(code, line)
    type_map = _infer_type_map(code)
    bindings = _infer_constant_bindings(code, line, ctx_data)
    receiver, attr_name = _attribute_at_position(line_text, character)
    if receiver and attr_name:
        inferred = _resolve_type_name(receiver, code, type_map)
        attrs, methods = _type_meta(inferred)
        if attr_name in attrs:
            contents = f"`{inferred}().{attr_name}`"
            return types.Hover(contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=contents))
        if attr_name in methods:
            contents = f"`{inferred}().{attr_name}()`"
            return types.Hover(contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=contents))

    word, _, _ = _word_at_position(line_text, character)
    if not word:
        return None
    if word in type_map:
        contents = f"`{word}` type: `{type_map[word]}()`"
        return types.Hover(contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=contents))
    if word in sigs:
        sig = sigs[word]
        contents = f"```avrae\n{sig.label}\n```\n\n{sig.doc}"
        return types.Hover(contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=contents))
    if word in bindings:
        return _format_binding_hover(word, bindings[word], "local")

    vars_map = ctx_data.vars.to_initial_names()
    if word in vars_map:
        return _format_binding_hover(word, vars_map[word], "var")

    gvars = resolver.snapshot()
    if word in gvars:
        return _format_binding_hover(word, gvars[word], "gvar")
    return None


def _current_prefix(line_text: str) -> str:
    match = IDENT_RE.search(line_text)
    return match.group(0) if match else ""


def _word_from_line(text: str, cursor: int) -> str:
    return _word_at_position(text, cursor)[0]


def _word_at_position(text: str, cursor: int) -> tuple[str, int, int]:
    cursor = max(0, min(cursor, len(text)))
    start = cursor
    while start > 0 and (text[start - 1].isalnum() or text[start - 1] == "_"):
        start -= 1
    end = cursor
    while end < len(text) and (text[end].isalnum() or text[end] == "_"):
        end += 1
    return text[start:end], start, end


def _line_text_to_cursor(code: str, line: int, character: int) -> str:
    lines = code.splitlines()
    if line >= len(lines):
        return ""
    return lines[line][:character]


def _sanitize_incomplete_line(code: str, line: int, character: int) -> str:
    lines = code.splitlines()
    if 0 <= line < len(lines):
        prefix = lines[line][:character].rstrip()
        if prefix.endswith("."):
            prefix = prefix[:-1]
        lines[line] = prefix
    return "\n".join(lines)


def _line_text(code: str, line: int) -> str:
    lines = code.splitlines()
    if line < 0 or line >= len(lines):
        return ""
    return lines[line]


def _infer_receiver_type(code: str, name: str) -> Optional[str]:
    return _infer_type_map(code).get(name)


def _infer_type_map(code: str) -> Dict[str, str]:
    try:
        tree = ast.parse(code)
    except SyntaxError:
        return {}
    type_map: dict[str, str] = {}

    class Visitor(ast.NodeVisitor):
        def visit_Assign(self, node: ast.Assign):
            val_type = self._value_type(node.value)
            if val_type:
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        type_map[target.id] = val_type
            self.generic_visit(node)

        def visit_AnnAssign(self, node: ast.AnnAssign):
            val_type = self._value_type(node.value) if node.value else None
            if val_type and isinstance(node.target, ast.Name):
                type_map[node.target.id] = val_type
            self.generic_visit(node)

        def _value_type(self, value: ast.AST | None) -> Optional[str]:
            if isinstance(value, ast.Call) and isinstance(value.func, ast.Name):
                if value.func.id in {"character", "combat"}:
                    return value.func.id
                if value.func.id == "vroll":
                    return "SimpleRollResult"
            if isinstance(value, ast.Name):
                if value.id in type_map:
                    return type_map[value.id]
                if value.id in {"character", "combat", "ctx"}:
                    return value.id
            if isinstance(value, ast.Attribute):
                attr_name = value.attr
                base_type = None
                if isinstance(value.value, ast.Name):
                    base_type = type_map.get(value.value.id)
                if base_type is None:
                    base_type = self._value_type(value.value)
                if base_type and attr_name in TYPE_MAP:
                    return attr_name
                return None
            return None

    Visitor().visit(tree)
    return type_map


def _resolve_type_name(receiver: str, code: str, type_map: Dict[str, str] | None = None) -> str:
    receiver = receiver.rstrip("()")
    mapping = type_map or _infer_type_map(code)
    if receiver in mapping:
        return mapping[receiver]
    if receiver in TYPE_MAP:
        return receiver
    tail = receiver.split(".")[-1]
    if tail in TYPE_MAP:
        return tail
    return receiver


def _type_meta(type_name: str) -> tuple[List[str], List[str]]:
    cls = TYPE_MAP.get(type_name)
    if cls is None:
        return [], []
    attrs = list(getattr(cls, "ATTRS", []))
    methods = list(getattr(cls, "METHODS", []))
    return sorted(set(attrs)), sorted(set(methods))


def _attribute_at_position(line_text: str, cursor: int) -> tuple[Optional[str], Optional[str]]:
    cursor = max(0, min(cursor, len(line_text)))
    for match in ATTR_AT_CURSOR_RE.finditer(line_text):
        start, end = match.span(2)
        if start <= cursor <= end:
            return match.group(1), match.group(2)
    return None, None


def _infer_constant_bindings(code: str, upto_line: int | None, ctx_data: ContextData) -> Dict[str, Any]:
    try:
        tree = ast.parse(code)
    except SyntaxError:
        return {}
    bindings: dict[str, Any] = {}
    limit = None if upto_line is None else upto_line + 1

    def _value_for(node: ast.AST) -> Any | None:
        value = _literal_value(node)
        if value is None:
            value = _evaluated_value(node, ctx_data, bindings)
        return value

    def _loop_binding(node: ast.AST) -> Any | None:
        value = _value_for(node)
        if value is None:
            return _LoopVarBinding()
        try:
            iterator = iter(value)
        except TypeError:
            return _LoopVarBinding()
        try:
            return next(iterator)
        except StopIteration:
            return _LoopVarBinding()

    class Visitor(ast.NodeVisitor):
        def visit_Assign(self, node: ast.Assign):
            if limit is not None and node.lineno > limit:
                return
            value = _value_for(node.value)
            if value is None:
                self.generic_visit(node)
                return
            for name in _names_from_target(node.targets):
                bindings[name] = value

        def visit_AnnAssign(self, node: ast.AnnAssign):
            if limit is not None and node.lineno > limit:
                return
            if node.value is None:
                return
            value = _value_for(node.value)
            if value is None:
                self.generic_visit(node)
                return
            for name in _names_from_target([node.target]):
                bindings[name] = value

        def visit_For(self, node: ast.For):
            if limit is not None and node.lineno > limit:
                return
            loop_val = _loop_binding(node.iter)
            for name in _names_from_target([node.target]):
                bindings[name] = loop_val
            self.generic_visit(node)

        def visit_AsyncFor(self, node: ast.AsyncFor):
            if limit is not None and node.lineno > limit:
                return
            loop_val = _loop_binding(node.iter)
            for name in _names_from_target([node.target]):
                bindings[name] = loop_val
            self.generic_visit(node)

    Visitor().visit(tree)
    return bindings


def _names_from_target(targets: Iterable[ast.expr]) -> List[str]:
    names: list[str] = []
    for target in targets:
        if isinstance(target, ast.Name):
            names.append(target.id)
        elif isinstance(target, (ast.Tuple, ast.List)):
            names.extend(_names_from_target(target.elts))
    return names


def _literal_value(node: ast.AST) -> Any | None:
    if isinstance(node, ast.Constant):
        return node.value
    if isinstance(node, ast.UnaryOp) and isinstance(node.op, (ast.UAdd, ast.USub)):
        val = _literal_value(node.operand)
        if isinstance(val, (int, float, complex)):
            return val if isinstance(node.op, ast.UAdd) else -val
        return None
    if isinstance(node, (ast.List, ast.Tuple, ast.Set)):
        items = []
        for elt in node.elts:
            val = _literal_value(elt)
            if val is None:
                return None
            items.append(val)
        if isinstance(node, ast.List):
            return items
        if isinstance(node, ast.Tuple):
            return tuple(items)
        return set(items)
    if isinstance(node, ast.Dict):
        keys = []
        values = []
        for k, v in zip(node.keys, node.values):
            key_val = _literal_value(k) if k is not None else None
            val_val = _literal_value(v)
            if key_val is None or val_val is None:
                return None
            keys.append(key_val)
            values.append(val_val)
        return dict(zip(keys, values))
    return None


def _evaluated_value(node: ast.AST, ctx_data: ContextData, bindings: Dict[str, Any] | None = None) -> Any | None:
    bindings = bindings or {}
    try:
        return _eval_node(node, ctx_data, bindings)
    except Exception:
        return None


def _eval_node(node: ast.AST, ctx_data: ContextData, bindings: Dict[str, Any]) -> Any | None:
    if isinstance(node, ast.Attribute):
        base = _eval_node(node.value, ctx_data, bindings)
        if base is None:
            return None
        return getattr(base, node.attr, None)
    if isinstance(node, ast.Call):
        if isinstance(node.func, ast.Name):
            if node.func.id == "character":
                return CharacterAPI(ctx_data.character)
            if node.func.id == "combat":
                return SimpleCombat(ctx_data.combat)
            if node.func.id == "range":
                args = []
                for arg in node.args:
                    val = _literal_value(arg)
                    if val is None:
                        return None
                    args.append(val)
                try:
                    return range(*args)
                except Exception:
                    return None
        if isinstance(node.func, ast.Attribute):
            base = _eval_node(node.func.value, ctx_data, bindings)
            if base is None:
                return None
            method_name = node.func.attr
            if not _is_safe_call(base, method_name):
                return None
            args = []
            for arg in node.args:
                val = _literal_value(arg)
                if val is None:
                    val = _eval_node(arg, ctx_data, bindings)
                if val is None:
                    return None
                args.append(val)
            kwargs = {}
            for kw in node.keywords:
                if kw.arg is None:
                    return None
                val = _literal_value(kw.value)
                if val is None:
                    val = _eval_node(kw.value, ctx_data, bindings)
                if val is None:
                    return None
                kwargs[kw.arg] = val
            callee = getattr(base, method_name, None)
            if not callable(callee):
                return None
            try:
                return callee(*args, **kwargs)
            except Exception:
                return None
    if isinstance(node, ast.Name):
        if node.id in bindings:
            return bindings[node.id]
        if node.id == "ctx":
            return AliasContextAPI(ctx_data.ctx)
    return None


def _is_safe_call(base: Any, method: str) -> bool:
    for cls, allowed in SAFE_METHODS.items():
        if isinstance(base, cls) and method in allowed:
            return True
    return False


def _format_binding_hover(name: str, value: Any, label: str) -> types.Hover:
    type_name = type(value).__name__
    preview = _preview_value(value)
    contents = f"**{label}** `{name}`\n\nType: `{type_name}`\nValue: `{preview}`"
    return types.Hover(contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=contents))


def _preview_value(value: Any) -> str:
    text = repr(value)
    if len(text) > 120:
        text = text[:117] + "..."
    return text


class _LoopVarBinding:
    def __repr__(self) -> str:
        return "<loop item>"
SAFE_METHODS: dict[type, set[str]] = {
    AliasLevels: {"get"},
    dict: {"get"},
}



import pytest

from avrae_ls.config import VarSources
from avrae_ls.context import ContextData, GVarResolver
from avrae_ls.runtime import FunctionRequiresCharacter, MockExecutor


def _ctx():
    return ContextData(vars=VarSources())


def _ctx_with_character(data: dict | None = None):
    return ContextData(character=data or {"name": "Aelar"}, vars=VarSources())


def _resolver(tmp_path):
    from avrae_ls.config import AvraeLSConfig

    cfg = AvraeLSConfig.default(tmp_path)
    res = GVarResolver(cfg)
    res.reset({"foo": "bar"})
    return res


@pytest.mark.asyncio
async def test_character_requires_active_character(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)

    result = await executor.run("character()", ctx, resolver)
    assert result.error is not None
    assert isinstance(getattr(result.error, "__cause__", None), FunctionRequiresCharacter)


@pytest.mark.asyncio
async def test_mock_executor_runs_code(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)

    result = await executor.run("x = 1\nx + 2", ctx, resolver)
    assert result.error is None
    assert result.value == 3
    assert result.stdout == ""


@pytest.mark.asyncio
async def test_mock_executor_resolves_gvars(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)

    result = await executor.run("get_gvar('foo')", ctx, resolver)
    assert result.error is None
    assert result.value == "bar"


@pytest.mark.asyncio
async def test_alias_block_executes(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    from avrae_ls.parser import primary_block_or_source

    alias_text = "!alias hello echo\n<drac2>\nx = 3\nreturn x\n</drac2>"
    code, _, _ = primary_block_or_source(alias_text)
    result = await executor.run(code, ctx, resolver)
    assert result.error is None
    assert result.value == 3


@pytest.mark.asyncio
async def test_argparse_available(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    result = await executor.run("pa = argparse('one two'); pa.get('one')", ctx, resolver)
    assert result.error is None


@pytest.mark.asyncio
async def test_err_raises_alias_exception(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    result = await executor.run("err('boom')", ctx, resolver)
    assert result.error is not None
    assert "boom" in str(result.error)


@pytest.mark.asyncio
async def test_consumables_list_safe(tmp_path):
    executor = MockExecutor()
    ctx = ContextData(character={"consumables": {"Ki": {"name": "Ki", "value": 3, "max": 5}}}, vars=VarSources())
    resolver = _resolver(tmp_path)
    result = await executor.run("x=character(); x.consumables", ctx, resolver)
    assert result.error is None
    assert result.value is not None
    first = list(result.value)[0]
    assert getattr(first, "name") == "Ki"


@pytest.mark.asyncio
async def test_actions_list_safe(tmp_path):
    executor = MockExecutor()
    ctx = ContextData(
        character={
            "actions": [
                {"name": "Second Wind", "activation_type": 3, "activation_type_name": "BONUS_ACTION", "description": "+1d10+5"}
            ]
        },
        vars=VarSources(),
    )
    resolver = _resolver(tmp_path)
    result = await executor.run("x=character(); x.actions", ctx, resolver)
    assert result.error is None
    actions = list(result.value)
    assert actions and actions[0].name == "Second Wind"


@pytest.mark.asyncio
async def test_character_attributes_accessible(tmp_path):
    executor = MockExecutor()
    ctx = ContextData(
        character={
            "name": "A",
            "stats": {"strength": 10, "dexterity": 10, "constitution": 10, "intelligence": 10, "wisdom": 10, "charisma": 10},
            "levels": {"Fighter": 1},
            "attacks": [{"name": "Punch"}],
            "skills": {"athletics": {"value": 2}},
            "saves": {"str": 2},
            "resistances": {"resist": [], "vuln": [], "immune": [], "neutral": []},
            "ac": 10,
            "max_hp": 10,
            "hp": 10,
            "temp_hp": 0,
            "spellbook": {"spells": []},
            "creature_type": "humanoid",
            "actions": [{"name": "Action"}],
            "coinpurse": {"gp": 1},
            "csettings": {},
            "race": "Human",
            "background": "Soldier",
            "owner": 1,
            "upstream": "up",
            "cvars": {"foo": "bar"},
            "consumables": {"Ki": {"name": "Ki", "value": 1, "max": 1}},
            "death_saves": {"successes": 0, "fails": 0},
            "description": "",
            "image": "",
        },
        vars=VarSources(),
    )
    resolver = _resolver(tmp_path)
    code = "\n".join(
        [
            "x=character()",
            "x.name",
            "x.stats",
            "x.levels",
            "x.attacks",
            "x.skills",
            "x.saves",
            "x.resistances",
            "x.ac",
            "x.max_hp",
            "x.hp",
            "x.temp_hp",
            "x.spellbook",
            "x.creature_type",
            "x.actions",
            "x.coinpurse",
            "x.csettings",
            "x.race",
            "x.background",
            "x.owner",
            "x.upstream",
            "x.cvars",
            "x.consumables",
            "x.death_saves",
            "x.description",
            "x.image",
        ]
    )
    result = await executor.run(code, ctx, resolver)
    assert result.error is None


@pytest.mark.asyncio
async def test_combat_returns_none_when_missing(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)

    result = await executor.run("combat()", ctx, resolver)
    assert result.error is None
    assert result.value is None


@pytest.mark.asyncio
async def test_combat_combatants_executes(tmp_path):
    from avrae_ls.config import AvraeLSConfig
    from avrae_ls.context import ContextBuilder

    cfg = AvraeLSConfig.default(tmp_path)
    ctx = ContextBuilder(cfg).build()
    resolver = _resolver(tmp_path)
    executor = MockExecutor()

    result = await executor.run("combat().combatants", ctx, resolver)
    assert result.error is None
    from collections.abc import Sequence

    assert isinstance(result.value, Sequence)
    assert len(result.value) == 2


class _EnsuringResolver(GVarResolver):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.calls: list[str] = []

    async def ensure(self, key: str) -> bool:
        self.calls.append(str(key))
        self._cache[str(key)] = f"fetched-{key}"
        return True


@pytest.mark.asyncio
async def test_get_gvar_prefetches_literal(tmp_path):
    from avrae_ls.config import AvraeLSConfig

    cfg = AvraeLSConfig.default(tmp_path)
    resolver = _EnsuringResolver(cfg)
    ctx = _ctx()
    executor = MockExecutor()

    result = await executor.run("get_gvar('abc123')", ctx, resolver)
    assert result.error is None
    assert resolver.calls == ["abc123"]
    assert result.value == "fetched-abc123"


@pytest.mark.asyncio
async def test_uvar_helpers_available(tmp_path):
    executor = MockExecutor()
    ctx = ContextData(vars=VarSources(uvars={"foo": "orig"}))
    resolver = _resolver(tmp_path)
    code = "\n".join(
        [
            "set_uvar_nx('foo', 'new')",
            "set_uvar_nx('bar', 2)",
            "set_uvar('baz', 3)",
            "get_uvars()",
        ]
    )

    result = await executor.run(code, ctx, resolver)
    assert result.error is None
    uvars = result.value
    assert uvars["foo"] == "orig"
    assert uvars["bar"] == 2
    assert uvars["baz"] == 3


@pytest.mark.asyncio
async def test_set_cvar_binds_runtime_name(tmp_path):
    executor = MockExecutor()
    ctx = _ctx_with_character()
    resolver = _resolver(tmp_path)
    result = await executor.run("c = character(); c.set_cvar('pet', 'hawk'); pet", ctx, resolver)
    assert result.error is None
    assert result.value == "hawk"


@pytest.mark.asyncio
async def test_reference_helpers_available(tmp_path):
    executor = MockExecutor()
    ctx = ContextData(vars=VarSources(cvars={"foo": 1}))
    resolver = _resolver(tmp_path)
    code = "\n".join(
        [
            "exists('foo')",
            "get('foo')",
            "typeof(123)",
            "ceil(2.1)",
            "floor(2.9)",
            "sqrt(9)",
            "round(2.51, 1)",
            "randchoice([1,2,3])",
            "load_json(dump_json({'a': 1}))['a']",
            "parse_coins('1')",
            "using(mod='foo')",
            "mod",
        ]
    )
    result = await executor.run(code, ctx, resolver)
    assert result.error is None
    assert isinstance(result.value, str)


def test_documented_builtins_present():
    executor = MockExecutor()
    ctx = _ctx()
    names = executor.available_names(ctx)
    documented = {
        "abs",
        "all",
        "any",
        "ceil",
        "enumerate",
        "float",
        "floor",
        "int",
        "len",
        "max",
        "min",
        "range",
        "round",
        "str",
        "sum",
        "time",
        "sqrt",
    }
    assert documented.issubset(names)



from __future__ import annotations

import io
import logging
import time
import ast
import json
import random
import math
try:  # optional dependency
    import yaml
except ImportError:  # pragma: no cover - fallback when PyYAML is absent
    yaml = None  # type: ignore
from dataclasses import dataclass
from typing import Any, Dict, Set, Callable

import d20
import draconic
from draconic.interpreter import _Break, _Continue, _Return

from .context import ContextData, GVarResolver
from .config import VarSources
from .api import AliasContextAPI, CharacterAPI, SimpleCombat, SimpleRollResult
from . import argparser as avrae_argparser
# Minimal stand-in for Avrae's AliasException
class AliasException(Exception):
    def __init__(self, msg, pm_user):
        super().__init__(msg)
        self.pm_user = pm_user


try:
    from avrae.aliasing.errors import FunctionRequiresCharacter  # type: ignore
except Exception:  # pragma: no cover - fallback when avrae is unavailable
    class FunctionRequiresCharacter(Exception):
        def __init__(self, msg: str | None = None):
            super().__init__(msg or "This alias requires an active character.")

log = logging.getLogger(__name__)


class MockNamespace:
    """A minimal attribute-friendly namespace used for ctx/combat/character."""

    def __init__(self, data: Dict[str, Any] | None = None):
        self._data = data or {}

    def __getattr__(self, item: str) -> Any:
        return self._data.get(item)

    def __getitem__(self, item: str) -> Any:
        return self._data.get(item)

    def __repr__(self) -> str:  # pragma: no cover - debugging helper
        return f"MockNamespace({self._data})"

    def to_dict(self) -> Dict[str, Any]:
        return dict(self._data)


@dataclass
class ExecutionResult:
    stdout: str
    value: Any = None
    error: BaseException | None = None


def _roll_dice(dice: str) -> int:
    roller = d20.Roller()
    try:
        result = roller.roll(str(dice))
    except d20.RollError:
        return 0
    return result.total


def _vroll_dice(dice: str, multiply: int = 1, add: int = 0) -> SimpleRollResult | None:
    roller = d20.Roller()
    try:
        dice_ast = roller.parse(str(dice))
    except d20.RollError:
        return None

    if multiply != 1 or add != 0:
        def _scale(node):
            if isinstance(node, d20.ast.Dice):
                node.num = (node.num * multiply) + add
            return node

        dice_ast = d20.utils.tree_map(_scale, dice_ast)

    try:
        rolled = roller.roll(dice_ast)
    except d20.RollError:
        return None
    return SimpleRollResult(rolled)


def _parse_coins(args: str):
    try:
        from avrae.aliasing.api.functions import parse_coins as avrae_parse_coins
    except Exception:
        avrae_parse_coins = None

    if avrae_parse_coins:
        try:
            return avrae_parse_coins(str(args))
        except Exception:
            pass

    # Fallback: accept numeric as gp, otherwise empty mapping.
    try:
        gp = float(str(args))
        return {"pp": 0, "gp": gp, "ep": 0, "sp": 0, "cp": 0, "total": gp}
    except Exception:
        return {"pp": 0, "gp": 0, "ep": 0, "sp": 0, "cp": 0, "total": 0}


def _default_builtins() -> Dict[str, Any]:
    return {
        "len": len,
        "min": min,
        "max": max,
        "sum": sum,
        "any": any,
        "all": all,
        "abs": abs,
        "range": range,
        "enumerate": enumerate,
        "sorted": sorted,
        "reversed": reversed,
        "int": int,
        "float": float,
        "str": str,
        "bool": bool,
        "round": round,
        "ceil": math.ceil,
        "floor": math.floor,
        "sqrt": math.sqrt,
        "time": time.time,
        "roll": _roll_dice,
        "vroll": _vroll_dice,
        "rand": random.random,
        "randint": random.randrange,
        "randchoice": random.choice,
        "randchoices": random.choices,
        "typeof": lambda inst: type(inst).__name__,
        "parse_coins": _parse_coins,
        "load_json": lambda s: json.loads(str(s)),
        "dump_json": lambda obj: json.dumps(obj),
        "load_yaml": lambda s: yaml.safe_load(str(s)) if yaml else None,
        "dump_yaml": (
            (lambda obj, indent=2: yaml.safe_dump(obj, indent=indent, sort_keys=False)) if yaml else (lambda obj, indent=2: str(obj))
        ),
    }


class MockExecutor:
    def __init__(self):
        self._base_builtins = _default_builtins()

    def available_names(self, ctx_data: ContextData) -> Set[str]:
        builtin_names = set(self._base_builtins.keys())
        runtime_names = {
            "ctx",
            "combat",
            "character",
            "roll",
            "vroll",
            "rand",
            "randint",
            "randchoice",
            "randchoices",
            "typeof",
            "parse_coins",
            "load_json",
            "dump_json",
            "load_yaml",
            "dump_yaml",
            "get_gvar",
            "get_svar",
            "get_cvar",
            "get_uvar",
            "get_uvars",
            "set_uvar",
            "set_uvar_nx",
            "delete_uvar",
            "uvar_exists",
            "print",
            "argparse",
            "err",
            "exists",
            "get",
            "using",
            "signature",
            "verify_signature",
        }
        variable_names = set(ctx_data.vars.to_initial_names().keys())
        return builtin_names | runtime_names | variable_names

    async def run(
        self,
        code: str,
        ctx_data: ContextData,
        gvar_resolver: GVarResolver | None = None,
    ) -> ExecutionResult:
        buffer = io.StringIO()
        resolver = gvar_resolver
        interpreter_ref: dict[str, draconic.DraconicInterpreter | None] = {"interpreter": None}
        runtime_character: CharacterAPI | None = None

        def _character_provider() -> CharacterAPI:
            nonlocal runtime_character
            interp = interpreter_ref["interpreter"]
            if not ctx_data.character:
                raise FunctionRequiresCharacter()
            if runtime_character is None and interp is not None:
                runtime_character = _RuntimeCharacter(ctx_data.character, ctx_data.vars, interp)
            if runtime_character is None:
                runtime_character = CharacterAPI(ctx_data.character)
            return runtime_character  # type: ignore[return-value]

        builtins = self._build_builtins(
            ctx_data,
            resolver,
            buffer,
            character_provider=_character_provider,
            interpreter_ref=interpreter_ref,
        )
        interpreter = draconic.DraconicInterpreter(
            builtins=builtins,
            initial_names=ctx_data.vars.to_initial_names(),
        )
        interpreter_ref["interpreter"] = interpreter

        value = None
        error: BaseException | None = None
        code_to_run = code
        try:
            parsed = interpreter.parse(code_to_run)
        except BaseException:
            wrapped, _ = _wrap_draconic(code_to_run)
            code_to_run = wrapped
            try:
                parsed = interpreter.parse(code_to_run)
            except BaseException as exc:
                error = exc
                log.debug("Mock execution error: %s", exc, exc_info=exc)
                return ExecutionResult(stdout=buffer.getvalue(), value=value, error=error)

        if resolver:
            await _ensure_literal_gvars(code_to_run, resolver)

        try:
            interpreter._preflight()
            value = self._exec_with_value(interpreter, parsed)
        except BaseException as exc:  # draconic raises BaseException subclasses
            error = exc
            log.debug("Mock execution error: %s", exc, exc_info=exc)
        return ExecutionResult(stdout=buffer.getvalue(), value=value, error=error)

    def _build_builtins(
        self,
        ctx_data: ContextData,
        resolver: GVarResolver | None,
        buffer: io.StringIO,
        character_provider: Callable[[], CharacterAPI] | None = None,
        interpreter_ref: Dict[str, draconic.DraconicInterpreter | None] | None = None,
    ) -> Dict[str, Any]:
        builtins = dict(self._base_builtins)
        var_store = ctx_data.vars
        interpreter_ref = interpreter_ref or {"interpreter": None}

        def _print(*args, sep=" ", end="\n"):
            buffer.write(sep.join(map(str, args)) + end)

        def _get_gvar(address: str):
            if resolver is None:
                return None
            return resolver.get_local(address)

        def _get_svar(name: str, default=None):
            return var_store.svars.get(str(name), default)

        def _get_cvar(name: str, default=None):
            return var_store.cvars.get(str(name), default)

        def _get_uvar(name: str, default=None):
            return var_store.uvars.get(str(name), default)

        def _get_uvars():
            return dict(var_store.uvars)

        def _set_uvar(name: str, value: Any):
            var_store.uvars[str(name)] = value
            return value

        def _set_uvar_nx(name: str, value: Any):
            key = str(name)
            if key not in var_store.uvars:
                var_store.uvars[key] = value
            return var_store.uvars[key]

        def _delete_uvar(name: str):
            return var_store.uvars.pop(str(name), None)

        def _uvar_exists(name: str) -> bool:
            return str(name) in var_store.uvars

        def _exists(name: str) -> bool:
            interp = interpreter_ref.get("interpreter")
            if interp is None:
                return False
            return str(name) in getattr(interp, "_names", {})

        def _get(name: str, default=None):
            interp = interpreter_ref.get("interpreter")
            if interp is None:
                return default
            return getattr(interp, "_names", {}).get(str(name), default)

        def _using(**imports):
            interp = interpreter_ref.get("interpreter")
            if interp is None:
                return None
            for ns, addr in imports.items():
                val = None
                if resolver:
                    val = resolver.get_local(addr)
                interp._names[str(ns)] = val  # type: ignore[attr-defined]
            return None

        def _signature(data=0):
            return f"signature:{int(data)}"

        def _verify_signature(sig):
            try:
                return {"signature": str(sig), "valid": True}
            except Exception:
                return {"signature": None, "valid": False}

        def _argparse(args, character=None, splitter=avrae_argparser.argsplit, parse_ephem=True):
            return avrae_argparser.argparse(args, character=character, splitter=splitter, parse_ephem=parse_ephem)

        def _err(reason, pm_user: bool = False):
            raise AliasException(str(reason), pm_user)

        ns_ctx = AliasContextAPI(ctx_data.ctx)
        ns_combat = SimpleCombat(ctx_data.combat) if ctx_data.combat else None
        if character_provider:
            character_fn = character_provider
        else:
            ns_character = CharacterAPI(ctx_data.character) if ctx_data.character else None

            def character_fn():
                if ns_character is None:
                    raise FunctionRequiresCharacter()
                return ns_character

        builtins.update(
            print=_print,
            roll=_roll_dice,
            vroll=_vroll_dice,
            ctx=ns_ctx,
            combat=lambda: ns_combat,
            character=lambda: character_fn(),
            get_gvar=_get_gvar,
            get_svar=_get_svar,
            get_cvar=_get_cvar,
            get_uvar=_get_uvar,
            get_uvars=_get_uvars,
            set_uvar=_set_uvar,
            set_uvar_nx=_set_uvar_nx,
            delete_uvar=_delete_uvar,
            uvar_exists=_uvar_exists,
            argparse=_argparse,
            err=_err,
            exists=_exists,
            get=_get,
            using=_using,
            signature=_signature,
            verify_signature=_verify_signature,
        )
        return builtins

    def _exec_with_value(self, interpreter: draconic.DraconicInterpreter, body) -> Any:
        last_val = None
        for expression in body:
            retval = interpreter._eval(expression)  # type: ignore[attr-defined]
            if isinstance(retval, (_Break, _Continue)):
                raise draconic.DraconicSyntaxError.from_node(retval.node, msg="Loop control outside loop", expr=interpreter._expr)  # type: ignore[attr-defined]
            if isinstance(retval, _Return):
                return retval.value
            last_val = retval
        return last_val


class _RuntimeCharacter(CharacterAPI):
    """Character wrapper that keeps mock runtime bindings in sync with cvar mutations."""

    def __init__(self, data: Dict[str, Any], var_store: VarSources, interpreter: draconic.DraconicInterpreter):
        super().__init__(data)
        self._var_store = var_store
        self._interpreter = interpreter

    def set_cvar(self, name: str, val: Any) -> Any:
        bound_val = super().set_cvar(name, val)
        key = str(name)
        self._var_store.cvars[key] = bound_val
        try:
            # Mirror Avrae behavior: new cvars are available as locals immediately.
            self._interpreter._names[key] = bound_val  # type: ignore[attr-defined]
        except Exception:
            pass
        return bound_val

    def set_cvar_nx(self, name: str, val: Any) -> Any:
        key = str(name)
        if key in self._var_store.cvars:
            return self._var_store.cvars[key]
        return self.set_cvar(key, val)

    # delete_cvar intentionally does not unbind runtime names, matching Avrae's docs.


def _wrap_draconic(code: str) -> tuple[str, int]:
    indented = "\n".join(f"    {line}" for line in code.splitlines())
    wrapped = f"def __alias_main__():\n{indented}\n__alias_main__()"
    return wrapped, 1


async def _ensure_literal_gvars(code: str, resolver: GVarResolver) -> None:
    for key in _literal_gvars(code):
        try:
            await resolver.ensure(key)
        except Exception as exc:  # pragma: no cover - defensive
            log.debug("Failed to prefetch gvar %s: %s", key, exc)


def _literal_gvars(code: str) -> Set[str]:
    try:
        tree = ast.parse(code)
    except SyntaxError:
        wrapped, _ = _wrap_draconic(code)
        try:
            tree = ast.parse(wrapped)
        except SyntaxError:
            return set()

    gvars: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id == "get_gvar":
            if not node.args:
                continue
            arg = node.args[0]
            if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                gvars.add(arg.value)
            elif isinstance(arg, ast.Str):
                gvars.add(arg.s)
    return gvars



import pytest

from avrae_ls.alias_preview import render_alias_command
from avrae_ls.alias_preview import simulate_command, validate_embed_payload
from avrae_ls.config import VarSources
from avrae_ls.context import ContextData, GVarResolver
from avrae_ls.runtime import MockExecutor


def _ctx():
    return ContextData(vars=VarSources())


def _resolver(tmp_path):
    from avrae_ls.config import AvraeLSConfig

    cfg = AvraeLSConfig.default(tmp_path)
    res = GVarResolver(cfg)
    res.reset({})
    return res


@pytest.mark.asyncio
async def test_preview_preserves_newlines(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    alias_text = "!alias hello echo\n<drac2>\nx = 3\ny = 4\nreturn x + y\n</drac2>\nworld"

    rendered = await render_alias_command(alias_text, executor, ctx, resolver)
    assert rendered.command == "echo\n7\nworld"


@pytest.mark.asyncio
async def test_preview_plain_alias(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    alias_text = "!alias hello echo world!"

    rendered = await render_alias_command(alias_text, executor, ctx, resolver)
    assert rendered.command == "echo world!"


@pytest.mark.asyncio
async def test_preview_argument_parsing(tmp_path):
    executor = MockExecutor()
    ctx = _ctx()
    resolver = _resolver(tmp_path)
    alias_text = '!alias hello echo %1% <drac2>return "&1&"</drac2>'

    rendered = await render_alias_command(alias_text, executor, ctx, resolver, args=["first arg"])
    assert rendered.command.startswith('echo "first arg"')
    assert rendered.last_value == "first arg"


def test_simulate_command_accepts_embed():
    payload = '-title "Hello"'
    preview, name, validation = simulate_command(f"embed {payload}")
    assert name == "embed"
    assert preview == payload
    assert validation is None


def test_simulate_command_embeds_validate_unknown_keys():
    payload = "-foo 1"
    _preview, _name, validation = simulate_command(f"embed {payload}")
    assert validation
    ok, err = validate_embed_payload(payload)
    assert not ok
    assert "unknown flag" in err


def test_validate_embed_flag_style():
    payload = '-title "Hello" -desc "World" -color #ff00ff -t 120 -f "Donuts|I have 15 donuts|inline"'
    ok, err = validate_embed_payload(payload)
    assert ok
    assert err is None


def test_validate_embed_flag_unknown():
    ok, err = validate_embed_payload("-foo bar")
    assert not ok
    assert "unknown flag" in err


def test_validate_embed_flag_color():
    ok, err = validate_embed_payload("-color nothex")
    assert not ok
    assert "color must be a 6-hex" in err


def test_validate_embed_field_format():
    ok, err = validate_embed_payload('-f "BadField"')
    assert not ok
    assert "field must be" in err.lower()



from avrae_ls.argument_parsing import apply_argument_parsing


def test_argument_parsing_replacements():
    text = "echo %1% %*% <drac2>return &1& + len(&ARGS&)</drac2>"
    rendered = apply_argument_parsing(text, ["first arg", "second"])
    assert '"first arg"' in rendered  # %1% gets quoted for spaces
    assert "first arg second" in rendered  # %*% preserved order
    assert "[\'first arg\', \'second\']" in rendered or "[\"first arg\", \"second\"]" in rendered
    assert "&1&" not in rendered
    assert "&ARGS&" not in rendered



from __future__ import annotations

import ast
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

import draconic
from lsprotocol import types

from .argument_parsing import apply_argument_parsing
from .parser import find_draconic_blocks

log = logging.getLogger(__name__)


@dataclass
class SymbolEntry:
    name: str
    kind: types.SymbolKind
    range: types.Range
    selection_range: types.Range


class SymbolTable:
    def __init__(self, entries: List[SymbolEntry]):
        self._entries = entries
        self._index: Dict[str, SymbolEntry] = {entry.name: entry for entry in entries}

    @property
    def entries(self) -> List[SymbolEntry]:
        return self._entries

    def lookup(self, name: str) -> Optional[SymbolEntry]:
        return self._index.get(name)


def build_symbol_table(source: str) -> SymbolTable:
    entries: list[SymbolEntry] = []
    parsed_source = apply_argument_parsing(source)
    blocks = find_draconic_blocks(parsed_source)
    if not blocks:
        entries.extend(_symbols_from_code(parsed_source, 0))
    else:
        for block in blocks:
            entries.extend(_symbols_from_code(block.code, block.line_offset))
    return SymbolTable(entries)


def document_symbols(source: str) -> List[types.DocumentSymbol]:
    table = build_symbol_table(source)
    return [
        types.DocumentSymbol(
            name=entry.name,
            kind=entry.kind,
            range=entry.range,
            selection_range=entry.selection_range,
        )
        for entry in table.entries
    ]


def find_definition_range(table: SymbolTable, name: str) -> types.Range | None:
    entry = table.lookup(name)
    if entry:
        return entry.selection_range
    return None


def _symbols_from_code(code: str, line_offset: int) -> List[SymbolEntry]:
    parser = draconic.DraconicInterpreter()
    local_offset = line_offset
    try:
        body = parser.parse(code)
    except draconic.DraconicSyntaxError:
        wrapped, added = _wrap_draconic(code)
        try:
            body = parser.parse(wrapped)
            local_offset += -added
        except draconic.DraconicSyntaxError:
            return []
    except Exception as exc:  # pragma: no cover - defensive
        log.debug("Symbol extraction failed: %s", exc)
        return []

    entries: list[SymbolEntry] = []
    for node in body:
        entry = _entry_from_node(node, local_offset)
        if entry:
            entries.append(entry)
    return entries


def _entry_from_node(node: ast.AST, line_offset: int = 0) -> SymbolEntry | None:
    if isinstance(node, ast.FunctionDef):
        kind = types.SymbolKind.Function
        name = node.name
    elif isinstance(node, ast.ClassDef):
        kind = types.SymbolKind.Class
        name = node.name
    elif isinstance(node, ast.Assign) and node.targets:
        target = node.targets[0]
        if isinstance(target, ast.Name):
            kind = types.SymbolKind.Variable
            name = target.id
        else:
            return None
    else:
        return None

    rng = _range_from_positions(
        getattr(node, "lineno", 1),
        getattr(node, "col_offset", 0) + 1,
        getattr(node, "end_lineno", None),
        getattr(node, "end_col_offset", None),
    )
    rng = _shift_range(rng, line_offset)
    return SymbolEntry(name=name, kind=kind, range=rng, selection_range=rng)


def _range_from_positions(
    lineno: int | None,
    col_offset: int | None,
    end_lineno: int | None,
    end_col_offset: int | None,
) -> types.Range:
    start = types.Position(
        line=max((lineno or 1) - 1, 0),
        character=max((col_offset or 1) - 1, 0),
    )
    end = types.Position(
        line=max(((end_lineno or lineno or 1) - 1), 0),
        character=max(((end_col_offset or col_offset or 1) - 1), 0),
    )
    return types.Range(start=start, end=end)


def _shift_range(rng: types.Range, line_offset: int) -> types.Range:
    if line_offset == 0:
        return rng
    return types.Range(
        start=types.Position(line=max(rng.start.line + line_offset, 0), character=rng.start.character),
        end=types.Position(line=max(rng.end.line + line_offset, 0), character=rng.end.character),
    )


def _wrap_draconic(code: str) -> tuple[str, int]:
    indented = "\n".join(f"    {line}" for line in code.splitlines())
    wrapped = f"def __alias_main__():\n{indented}\n__alias_main__()"
    return wrapped, 1



from pathlib import Path

from avrae_ls.api import AliasContextAPI, CharacterAPI, CombatAPI
from avrae_ls.config import AvraeLSConfig


def _profile():
    cfg = AvraeLSConfig.default(Path("."))
    return cfg.profiles[cfg.default_profile]


def test_character_api_mock_schema():
    profile = _profile()
    ch = CharacterAPI(profile.character)
    assert ch.name == "Aelar Wyn"
    assert ch.stats.prof_bonus == 3
    assert ch.levels.total_level == 5
    assert ch.skills.athletics.value >= 5
    assert ch.saves.get("str").value >= 5
    assert ch.coinpurse.total > 0
    assert ch.cc_exists("Hit Dice")
    start = ch.get_cc("Hit Dice")
    ch.mod_cc("Hit Dice", -1)
    assert ch.get_cc("Hit Dice") == start - 1
    ch.reset_hp()
    assert ch.hp == ch.max_hp
    assert ch.spellbook.can_cast("Cure Wounds", 1)


def test_combat_api_mock_schema():
    profile = _profile()
    cmb = CombatAPI(profile.combat)
    assert cmb.round_num >= 1
    assert cmb.me is not None
    assert cmb.get_combatant(cmb.me.name) is not None
    assert cmb.get_group("Goblins") is not None
    assert cmb.get_metadata("scene") == "Forest road"
    cmb.set_metadata("note", "testing")
    assert cmb.get_metadata("note") == "testing"
    cmb.end_round()
    assert cmb.turn_num == 0


def test_ctx_api_mock_schema():
    profile = _profile()
    ctx = AliasContextAPI(profile.ctx)
    assert ctx.guild.name
    assert ctx.channel.name
    assert ctx.author.display_name
    assert len(ctx.author.get_roles()) >= 1



from __future__ import annotations

import d20


class RerollableStringifier(d20.SimpleStringifier):
    """Stringifier that emits rerollable expressions and skips dropped dice."""

    def _stringify(self, node):
        if not node.kept:
            return None
        return super()._stringify(node)

    def _str_expression(self, node):
        return self._stringify(node.roll)

    def _str_literal(self, node):
        return str(node.total)

    def _str_parenthetical(self, node):
        return f"({self._stringify(node.value)})"

    def _str_set(self, node):
        out = f"{', '.join([self._stringify(v) for v in node.values if v.kept])}"
        if len(node.values) == 1:
            return f"({out},)"
        return f"({out})"

    def _str_dice(self, node):
        return self._str_set(node)

    def _str_die(self, node):
        return str(node.total)



from pathlib import Path

from avrae_ls.completions import completion_items_for_position, hover_for_position
from avrae_ls.config import AvraeLSConfig
from avrae_ls.context import ContextData, GVarResolver


def test_hover_out_of_bounds_cursor_does_not_crash():
    cfg = AvraeLSConfig.default(Path("."))
    ctx_data = ContextData()
    resolver = GVarResolver(cfg)
    # cursor beyond line length should not raise
    hover = hover_for_position("character.name\n", 0, 999, {}, ctx_data, resolver)
    # optional hover may be None; the important part is no exception
    assert hover is None or hover.contents


def test_attribute_completion_from_variable_binding():
    code = "x = character()\ny = x.\n"
    items = completion_items_for_position(code, 1, len("y = x."), [])
    labels = {item.label for item in items}
    assert "levels" in labels
    assert "name" in labels


def test_nested_attributes_completion():
    code = "x = character()\ny = x.levels\nz = y."
    items = completion_items_for_position(code, line=2, character=len("z = y."), suggestions=[])
    labels = {item.label for item in items}
    assert len(labels) != 0


from pathlib import Path

from avrae_ls.cvars import derive_character_cvars
from avrae_ls.config import AvraeLSConfig, VarSources
from avrae_ls.context import ContextBuilder


def test_var_sources_scope_order():
    sources = VarSources(cvars={"foo": "c"}, uvars={"foo": "u"}, svars={"server": "s"})
    names = sources.to_initial_names()
    assert names["foo"] == "c"
    assert "server" not in names


def test_derive_character_cvars_matches_table():
    character = {
        "name": "Ally",
        "stats": {"strength": 15, "dexterity": 12, "constitution": 8, "intelligence": 10, "wisdom": 13, "charisma": 9, "prof_bonus": 3},
        "saves": {"str": 5, "dex": 3, "con": 1, "int": 0, "wis": 4, "cha": 2},
        "levels": {"Fighter": 2, "Wizard": 1},
        "ac": 17,
        "max_hp": 30,
        "spellbook": {"spell_mod": 4},
        "description": "Ranger of the north",
        "image": "https://example.invalid/ally.png",
        "csettings": {"color": 255},
    }

    cvars = derive_character_cvars(character)
    assert cvars["strength"] == 15
    assert cvars["strengthMod"] == 2
    assert cvars["strengthSave"] == 5
    assert cvars["dexterity"] == 12
    assert cvars["dexterityMod"] == 1
    assert cvars["dexteritySave"] == 3
    assert cvars["constitutionSave"] == 1
    assert cvars["charismaMod"] == -1
    assert cvars["armor"] == 17
    assert cvars["hp"] == 30
    assert cvars["level"] == 3
    assert cvars["FighterLevel"] == 2
    assert cvars["WizardLevel"] == 1
    assert cvars["proficiencyBonus"] == 3
    assert cvars["spell"] == 4
    assert cvars["description"] == "Ranger of the north"
    assert cvars["image"].startswith("https://example.invalid/")
    assert cvars["color"] == "ff"


def test_context_builder_binds_character_and_cvar_table(tmp_path: Path):
    cfg = AvraeLSConfig.default(tmp_path)
    cfg.profiles["default"].character["cvars"] = {"pet": "hawk"}
    cfg.profiles["default"].vars = VarSources.from_data({"uvars": {"timezone": "UTC"}, "svars": {"ignored": True}})

    ctx = ContextBuilder(cfg).build()
    names = ctx.vars.to_initial_names()

    assert names["pet"] == "hawk"
    assert names["timezone"] == "UTC"
    assert names["strength"] == cfg.profiles["default"].character["stats"]["strength"]
    assert "ignored" not in names



import pytest

from avrae_ls.config import AvraeLSConfig, DiagnosticSettings, VarSources
from avrae_ls.context import ContextData, GVarResolver
from avrae_ls.diagnostics import DiagnosticProvider
from avrae_ls.runtime import MockExecutor


def _provider():
    executor = MockExecutor()
    settings = DiagnosticSettings()
    return DiagnosticProvider(executor, settings)


def _resolver(tmp_path):
    cfg = AvraeLSConfig.default(tmp_path)
    return GVarResolver(cfg)


@pytest.mark.asyncio
async def test_reports_syntax_error(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("if True print('hi')", ctx_data, resolver)
    assert diags
    assert any(d.severity.value == 1 for d in diags)  # DiagnosticSeverity.Error


@pytest.mark.asyncio
async def test_reports_unknown_name(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("x + 1", ctx_data, resolver)
    assert any("undefined" in d.message for d in diags)


@pytest.mark.asyncio
async def test_reports_unknown_gvar(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("get_gvar('abc')", ctx_data, resolver)
    assert any("gvar" in d.message for d in diags)


@pytest.mark.asyncio
async def test_fetches_gvar_when_enabled(tmp_path):
    cfg = AvraeLSConfig.default(tmp_path)
    cfg.enable_gvar_fetch = True

    class _RecordingResolver(GVarResolver):
        def __init__(self, cfg):
            super().__init__(cfg)
            self.calls: list[str] = []

        async def ensure(self, key: str) -> bool:
            self.calls.append(str(key))
            self._cache[str(key)] = f"fetched-{key}"
            return True

    provider = _provider()
    resolver = _RecordingResolver(cfg)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("get_gvar('abc123')", ctx_data, resolver)
    assert not any("gvar" in d.message for d in diags)
    assert resolver.calls == ["abc123"]
    assert resolver.get_local("abc123") == "fetched-abc123"


@pytest.mark.asyncio
async def test_for_loop_binds_target(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    code = "for x in range(3):\n    y = x\nprint(x)"
    diags = await provider.analyze(code, ctx_data, resolver)
    assert not any("undefined" in d.message for d in diags)


@pytest.mark.asyncio
async def test_drac_block_for_loop_return(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    alias_text = "!alias aaa echo \n<drac2>\nfor i in range(3):\n  return i\n\n</drac2>"
    diags = await provider.analyze(alias_text, ctx_data, resolver)
    assert not any("undefined" in d.message for d in diags)


@pytest.mark.asyncio
async def test_handles_alias_drac_block(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    alias_text = "!alias hello echo\n<drac2>\nx = 3\nreturn x\n</drac2>"
    diags = await provider.analyze(alias_text, ctx_data, resolver)
    assert diags == []


@pytest.mark.asyncio
async def test_reports_bad_args(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("len(1, 2)", ctx_data, resolver)
    assert any("invalid arguments" in d.message for d in diags)


@pytest.mark.asyncio
async def test_reports_import_usage(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("import os\nx=1", ctx_data, resolver)
    assert any("Imports are not supported" in d.message for d in diags)


@pytest.mark.asyncio
async def test_blocks_private_method_calls(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("class X:\n    def _hidden(self):\n        return 1\n\nX()._hidden()", ctx_data, resolver)
    assert any("private methods" in d.message for d in diags)


@pytest.mark.asyncio
async def test_roll_and_vroll_available(tmp_path):
    provider = _provider()
    resolver = _resolver(tmp_path)
    ctx_data = ContextData(vars=VarSources())

    diags = await provider.analyze("x = roll('1d1')\ny = vroll('1d1')\nx + y.total", ctx_data, resolver)
    assert not any("undefined" in d.message for d in diags)
